#' Generation of Procscal Data
#'
#' @param ns # of matrices to be generated.
#' @param nv # of rows of the matrices to be rotated.
#' @param nr # of columns of the matrices to be rotated.
#' @param genm = 1 to generate m vectors, the column mean
#' @param genW = 1 to generate scalar W matrices, the left weight \cr
#' = 2  to generate diagonal W matrices, the left weight
#' @param genT = 0 to skip the generation of T matrices, the rotation
#' @param genW2 = 1 to generate W2 matrices, the right weigt
#' @param genmu = 1 to generate mu matrix
#' @param mu0  The common value of mu
#' @param genD1 = 1 to generate D1 matrices,
#' @param genD2 = 1 to generate D2 matrices,
#' @param eqm = 1 to generate common m
#' @param eqW = 1 to generate common W
#' @param eqT = 1 to generate common T
#' @param eqW2 = 1 to generate common W2
#' @param diagW2 = 0 to make W2 non-diagonal
#' @param eqD1 = 1 to generate common WD1
#' @param eqD2 = 1 to generate common D2
#' @param genV = 1 to generate V matrix
#' @param diagV = 0 to make V non-diagonal
#' @param pcA = 1 to make the column variances of A equal \cr
#' = 2 to column orthogonalize A
#' @param errstd  std of errors to be added to final C matrix
#' @param pmiss = proportion of missing elements of C matrix
#' @param minWD = minimum value of W and D matrices
#'
#' @details
#' This program generates A, mk, Tk, Wk, W2k, D1k, D2k matrices \cr
#' so that the k-th target matrix Ck is created as \cr
#'  Ck = inv(D1k)  ( A Wk Tk W2k + 1 mk' ) inv(D2k) + Ek, k=1,2,...,ns, \cr
#'  where Disp(Ek) = errstd^2*solve(V). \cr
#' \cr
#' When genmu=1, we set genm=0; genD1=0; genD2=0; genW2=0; genV=0 and \cr
#'  muk = mu0 + A uk \cr
#' will be generated via uk where muk is nv x 1 and uk is nr x 1. \cr
#' \cr
#' For examples, see procotWKS or procscal.
#'
#' @return  A list of all the matrices and the parameter values.
#'
#' @export
#'

gendataWmm <- function( ns, nv, nr
                      , genm=0, genW=0, genT=1, genW2=0, genmu=0, mu0=0
                      , genD1=0, genD2=0, eqm=0, eqW=0, eqT=0, eqW2=0, diagW2=1
                      , eqD1=0, eqD2=0, genV=0, diagV=1, pcA=0, errstd=0
                      , pmiss=0, minWD=0.2 ){
 # generation of data matrix for expanded Procrustes rotation
 # Shin-ichi Mayekawa
 # 111108,09
 # weight: 111220
 # multigroup: 111220
 # chage the storage order: 111221
 # D1 and D2: 111223
 # eq options: 111223
 # W2 weight: 111225,27
 # pcA: 120104
 # genV std: 120105
 # genmu: 121020
 # names: 130126
 # varC varE: 130127
 # W=1-(1-minWD)*runif(): 130129
 # bugfix: 130307
 #
 # A     is  nv x nr
 # C_k   is  nv x nr
 # W_k   is  nr x nr diagonal
 # T_k   is  nr x nr orthonormal
 # m_k   is  nr x 1
 # D1_k  is  nv x nv diagonal
 # D2_k  is  nr x nr diagonal
 #
 # For each k in 1:ns  generate
 #  Ck = inv(D1k)  ( A Wk Tk W2k + 1 mk' ) inv(D2k)
 # where A is nv x nr,  Wk is nr x nr diag, Tk is nr x nr orthonormal
 #       mk is nv x 1,  D1k is nv x nv diag,  and D2k is nr x nr diag.
 #
 # In addition, if genmu == 1,
 # we set   genm=0; genD1=0; genD2=0; genW2=0; genV=0
 # and
 #  muk = mu0 + A uk
 # will be generated where muk is nv x 1, uk is nr x 1 and mu0=0.
 #  ( uk is generated using random numbers. )
 #
 #
 # Note that, A is the only common component.
 #
 # Value: as list
 #
 # $C
 # $C[[1]]
 #            dim1        dim2
 # v1  -1.23931063 -0.05189077
 # v20 -2.42441958  0.71727658
 #
 # $C[[2]]
 #             dim1        dim2
 # v1   0.976908657 -0.76608693
 # v20  1.377557974 -2.11988357
 #
 # $C[[3]]
 #            dim1       dim2
 # v1  -1.15689070 -0.4474560
 # v20 -2.52788088 -0.0990135
 #
 #
 # $A
 #            dim1        dim2
 # v1   1.07623970  0.61749324
 # v20  2.48297560  0.48158535
 #
 # $m
 # $m[[1]]
 #      [,1]
 # dim1    0
 # dim2    0
 #
 # $m[[2]]
 #      [,1]
 # dim1    0
 # dim2    0
 #
 # $m[[3]]
 #      [,1]
 # dim1    0
 # dim2    0
 #
 #
 # $W
 # $W[[1]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 # $W[[2]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 # $W[[3]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 #
 # $T
 # $T[[1]]
 #            dim1       dim2
 # dim1 -0.8873268  0.4611411
 # dim2 -0.4611411 -0.8873268
 #
 # $T[[2]]
 #           dim1       dim2
 # dim1 0.3754233 -0.9268534
 # dim2 0.9268534  0.3754233
 #
 # $T[[3]]
 #            dim1       dim2
 # dim1 -0.9884336  0.1516546
 # dim2 -0.1516546 -0.9884336
 #
 #
 # $W2
 # $W2[[1]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 # $W2[[2]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 # $W2[[3]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 #
 # $mu
 # $mu[[1]]
 #            [,1]
 # v1  -0.49092493
 # v20 -0.87874857
 #
 # $mu[[2]]
 #            [,1]
 # v1   0.25465697
 # v20  0.84638756
 #
 # $mu[[3]]
 #            [,1]
 # v1   0.23626796
 # v20  0.03236101
 #
 #
 # $u
 # $u[[1]]
 #            [,1]
 # dim1 -0.3016979
 # dim2 -0.2691943
 #
 # $u[[2]]
 #            [,1]
 # dim1  0.3941195
 # dim2 -0.2745133
 #
 # $u[[3]]
 #             [,1]
 # dim1 -0.09242161
 # dim2  0.54370760
 #
 #
 # $D1
 # $D1[[1]]
 #     v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12 v13 v14 v15 v16 v17 v18 v19 v20
 # v1   1  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v2   0  1  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v3   0  0  1  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v4   0  0  0  1  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v5   0  0  0  0  1  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v6   0  0  0  0  0  1  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v7   0  0  0  0  0  0  1  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v8   0  0  0  0  0  0  0  1  0   0   0   0   0   0   0   0   0   0   0   0
 # v9   0  0  0  0  0  0  0  0  1   0   0   0   0   0   0   0   0   0   0   0
 # v10  0  0  0  0  0  0  0  0  0   1   0   0   0   0   0   0   0   0   0   0
 # v11  0  0  0  0  0  0  0  0  0   0   1   0   0   0   0   0   0   0   0   0
 # v12  0  0  0  0  0  0  0  0  0   0   0   1   0   0   0   0   0   0   0   0
 # v13  0  0  0  0  0  0  0  0  0   0   0   0   1   0   0   0   0   0   0   0
 # v14  0  0  0  0  0  0  0  0  0   0   0   0   0   1   0   0   0   0   0   0
 # v15  0  0  0  0  0  0  0  0  0   0   0   0   0   0   1   0   0   0   0   0
 # v16  0  0  0  0  0  0  0  0  0   0   0   0   0   0   0   1   0   0   0   0
 # v17  0  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   1   0   0   0
 # v18  0  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   1   0   0
 # v19  0  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   1   0
 # v20  0  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   1
 #
 # $D1[[2]]
 #     v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12 v13 v14 v15 v16 v17 v18 v19 v20
 # v1   1  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v20  0  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   1
 #
 # $D1[[3]]
 #     v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12 v13 v14 v15 v16 v17 v18 v19 v20
 # v1   1  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   0
 # v20  0  0  0  0  0  0  0  0  0   0   0   0   0   0   0   0   0   0   0   1
 #
 #
 # $D2
 # $D2[[1]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 # $D2[[2]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 # $D2[[3]]
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 #
 # $V
 #      dim1 dim2
 # dim1    1    0
 # dim2    0    1
 #
 # $genm
 # [1] 0
 #
 # $genW
 # [1] 0
 #
 # $genT
 # [1] 1
 #
 # $genW2
 # [1] 0
 #
 # $genmu
 # [1] 1
 #
 # $genD1
 # [1] 0
 #
 # $genD2
 # [1] 0
 #
 # $eqm
 # [1] 0
 #
 # $eqW
 # [1] 0
 #
 # $eqT
 # [1] 0
 #
 # $eqD1
 # [1] 0
 #
 # $eqD2
 # [1] 0
 #
 # $eqW2
 # [1] 0
 #
 # $diagW2
 # [1] 1
 #
 # $pcA
 # [1] 2
 #
 # $genV
 # [1] 0
 #
 # $diagV
 # [1] 1
 #
 # $errstd
 # [1] 0.0005
 #
 # $pmiss
 # [1] 0
 #
 # $varC
 # [1] 1.206694
 #
 # $varE
 # [1] 1.41453

 #
 #
 # To convert the list of C matrices to 3D array, use
 # Call=res[[1]]
 # CC=array(unlist(Call), dim = c(dim(Call[[1]]), length(Call)))
 #

 # common A
 A=matrix(rnorm(nv*nr),nv,nr)

 # column center A
 if( genm == 1 ){
  mA=matrix(colSums(A)/nv,nr)
  A=A-matrix(1,nv,1)%*%t(mA)
 }

 # normalize A
 if( pcA == 2 ){
  # t(A)%*%A=diag
  ev=eigen(t(A)%*%A);
  eva=matrix(ev$values,length(ev$values),1); eve=ev$vectors;
  Q=eve
  A=A%*%Q
 }
 else if( pcA == 1) {
  # diag(t(A)%*%A)=nv I
  DAinv=Diag(sqrt(nv/colSums(A*A)))
  A=A%*%DAinv
 }

 # change the sign if needed
 sig=apply( A>0, 2, sum )
 loc=which( sig < nv/2 )
 dd=rep(1,nr);  dd[loc]=-1; dd=diag(dd)
 A=A%*%dd



 Call=NULL
 Wall=NULL
 W2all=NULL
 Tall=NULL
 mall=NULL
 D1all=NULL
 D2all=NULL
 uall=NULL
 muall=NULL

 # parafac?
 if( genmu == 1 ){
  genm=0; genD1=0; genD2=0; genW2=0; genV=0

  # generate ug and mug
  u=t( matrix(rnorm(ns*nr),nr,ns) )
  u=t( u-matrix(colMeans(u),nrow(u),ncol(u),byrow=1) )
  rownames(u)=paste("dim",1:nr,sep="")
  for( k in 1:ns ){
   mu=A%*%u[,k] + rep(mu0,nv)
   rownames(mu)=paste("v",1:nv,sep="")
   uall=append( uall, list(u[,k,drop=0]) )
   muall=append( muall, list(mu) )
  }
 } # end of parafac



 # equal m, W, T, Ds, W2
 m0=matrix(rnorm(nr),nr,1)
 if( genW == 2 ){
  W0=Diag(1-(1-minWD)*runif(nr))
 } else if( genW == 1 ){
  W0=Diag(rep(1-(1-minWD)*runif(1),nr))
 } else W=diag(nr)
 if( diagW2 == 1 ) W20=Diag(rep(1-(1-minWD)*runif(1),nr))
 else { W20=matrix(runif(nr*nr),nr,nr); W20=t(W20)%*%W20 }
 D10=Diag(floor(10*runif(nv)+1))
 D20=Diag(floor(10*runif(nr)+1))
 dum=matrix(rnorm(nv*nr),nv,nr)
 T0=eigen(t(dum)%*%dum)[[2]]


 # generate V
 if( genV == 1 ){
  if( diagV == 1 ){ V=diag(runif(nr)); V=V/sqrt(sum(V*V)) }
  else { V=matrix(rnorm(nv*nr),nv); V=t(V)%*%V }
  V=solve(V)
 } else V=diag(nr)
 rownames(V)=paste("dim",1:nr,sep="")
 colnames(V)=paste("dim",1:nr,sep="")




 for( k in 1:ns ){

  # generate T
  if( genT == 1 ){
   if( eqT == 1 ) T=T0
   else {
    dum=matrix(rnorm(nv*nr),nv,nr)
    T=eigen(t(dum)%*%dum)[[2]]
   }
  }
  else T=diag(nr)

  # generate m
  if( genm == 1 ){
   if( eqm == 1 ) m=m0
   else  m=matrix(rnorm(nr),nr,1)
  } else m=matrix(rep(0,nr),nr)

  # generate W
  if( genW == 2 ){
   if( eqW == 1 ) W=W0
   else W=Diag(1-(1-minWD)*runif(nr))
  } else if( genW == 1 ){
   if( eqW == 1 ) W=W0
   else W=Diag(rep(1-(1-minWD)*runif(1),nr))
  } else W=diag(nr)
  locsmall=which( diag(W) < minWD )
  diag(W)[locsmall]=minWD

  # generate W2
  if( genW2 == 1  &&  eqW2 == 1 ) W2=W20
  else if( genW2 == 1 ){
   if( diagW2 == 1) W2=Diag(runif(nr))
   else if( diagW2 == 0) { W2=matrix(runif(nr*nr),nr,nr); W2=t(t(W2)%*%W2) }
  }
  else W2=diag(nr)
  locsmall=which( diag(W2) < minWD )
  diag(W2)[locsmall]=minWD

  # normalize W2
  if( genW >= 1  && genW2 == 1 ){
   dd=sqrt(sum(W2*W2))
   W2=W2/dd
   W=W*dd
  }

  # generate D1 and D2
  if( genD1 == 1 ){
   if( eqD1 == 1 ) D1=D10
   else D1=Diag(floor(3*runif(nv)+1))
  } else D1=diag(nv)
  if( genD2 == 1 ){
   if( eqD2 == 1 ) D2=D20
   else D2=Diag(floor(3*runif(nr)+1))
  } else D2=diag(nr)
  locsmall=which( diag(D1) < minWD )
  diag(D1)[locsmall]=minWD
  locsmall=which( diag(D2) < minWD )
  diag(D2)[locsmall]=minWD

  # normalize D1 and D2
  if( genW > 0  || genW2 > 0 ){
   if( genD1 == 1 ) {
    d=sqrt(sum(D1*D1)/nv)
    D1=D1/d
   }
   if( genD2 == 1 ) {
    d=sqrt(sum(D2*D2)/nr)
    D2=D2/d
   }
  }

  # target matrix
  C=A%*%W%*%T%*%W2 + matrix(rep(1,nv),nv,1)%*%t(m)


  # add error
  varC=var(c(C))
  varE=0
  if( errstd != 0 ){
   E=matrix(rnorm(nv*nr),nv,nr)
   if( genV == 1  &&  diagV == 1){
    E=E%*%sqrt(solve(V))
    E=E%*%T
   }
   varE=var(c(E))
   C=C+errstd*sqrt(varC/varE)*E
  }

  # D matrices
  C=solve(D1)%*%C%*%solve(D2)


  dum=matrix(runif(nv*nr),nv,nr)
  C[which(dum < pmiss)]=NA

  rownames(A)=paste("v",1:nv,sep="")
  colnames(A)=paste("dim",1:nr,sep="")
  rownames(C)=paste("v",1:nv,sep="")
  colnames(C)=paste("dim",1:nr,sep="")
  rownames(m)=paste("dim",1:nr,sep="")
  rownames(D1)=paste("v",1:nv,sep="")
  colnames(D1)=paste("v",1:nv,sep="")
  rownames(D2)=paste("dim",1:nr,sep="")
  colnames(D2)=paste("dim",1:nr,sep="")
  rownames(W2)=paste("dim",1:nr,sep="")
  colnames(W2)=paste("dim",1:nr,sep="")
  rownames(W)=paste("dim",1:nr,sep="")
  colnames(W)=paste("dim",1:nr,sep="")
  rownames(T)=paste("dim",1:nr,sep="")
  colnames(T)=paste("dim",1:nr,sep="")

  Call=append( Call, list(C) )
  Tall=append( Tall, list(T) )
  Wall=append( Wall, list(W) )
  W2all=append( W2all, list(W2) )
  mall=append( mall, list(m) )
  D1all=append( D1all, list(D1) )
  D2all=append( D2all, list(D2) )


 } # end of k loop


 if( ns == 1 ){
  Call=Call[[1]]; Wall=Wall[[1]]; Tall=Tall[[1]]; mall=mall[[1]]
  D1all=D1all[[1]]; D2all=D2all[[1]]; W2all=W2all[[1]];
 }

 res=list( C=Call, A=A, m=mall, W=Wall, T=Tall, W2=W2all, mu=muall, u=uall
           , D1=D1all, D2=D2all, V=V
           , genm=genm, genW=genW, genT=genT, genW2=genW2, genmu=genmu
           , genD1=genD1, genD2=genD2, eqm=eqm, eqW=eqW, eqT=eqT
           , eqD1=eqD1, eqD2=eqD2, eqW2=eqW2, diagW2=diagW2
           , pcA=pcA, genV=genV, diagV=diagV, errstd=errstd
           , pmiss=pmiss, varC=varC, varE=varE )
 return( res )


} # of gendataWmm

