#' Generate Simple Structure Factor Pattern
#'
#' @param n # of variables
#' @param ndim # of dimensions
#' @param large The mean of large factor loadings
#' @param small The mean of small factor loadings
#' @param sd The standard deviation of factor loadings
#' @param pc = 1 to rotate the result so that A'A=diag. \cr
#' = 0 to do nothing.
#' @param reorder = 1 to reorder the rows by reorder_fa. \cr
#' = 2 to reorder the rows and columns by reorder_fa.
#'
#' @details
#' First, an
#' Independent Cluster factor loadings matrix, A0, which is a
#' n x ndim block diagonal rectangular matrix consisting of 1s and 0s,
#'  will be generated. \cr
#' Then, the 1s will be replaced by the random variables from
#' N(large,sd^2), and 0s, by the ones from N(small,sd^2). \cr
#' If pc=1, the above loadings matrix A will be rotated so that
#' \code{t(A)\%*\%A=I}. \cr
#' If reorder=1 or 2, the above A matrix will be reordered. \cr
#' \cr
#' Therefore, if \cr
#' \code{resg=gendatafa_A(n,ndim, pc=1, reorder=2)}
#' \cr
#' \code{(resg$loadings01\%*\%resg$rotmat)[resg$row_order,resg$col_order]}
#' \cr
#' make resg$loadings and A0 comparable.
#'
#' \code{large=1, small=0,  sd=0, pc=0, reorder=0} will create
#' a 0-1 block diagonal matrix.
#'
#' @return
#' A list of \cr
#' loadings = A matrix \cr
#' loadings01 = A0 matrix \cr
#' rotmat = The rotation matrix to pc \cr
#' row_order = The row reorder index \cr
#' col_order = The col reorder index
#'
#'
#' @examples
#' A01=gendatafa_A( 20, 4, large=1,small=0, pc=0, sd=0, reorder=0 )
#' A=gendatafa_A( 20, 4 )
#'
#' @export
#'

gendatafa_A <- function( n, ndim, large=.8, small=.2, sd=0.07, pc=0
                         , reorder=0 ){
 # generate simple structure factor pattern matrix
 # Shin-ichi Mayekawa
 # 20151204
 # reorder: 20151218
 # list output: 20151223
 # names: 20151224
 #

 # location of large loadings
 nn=rep( trunc(n/ndim), ndim )
 if( sum(nn) < n ){
  for( k in 1:ndim ){
   nn[ndim-k+1]=nn[ndim-k+1]+1
   if( sum(nn) == n ) break
  }
 }
 #Print(n,ndim,nn)

 # 0-1 block diagonal matrix
 A=diag(ndim)
 A=repmat( A, nn )
 dimnames(A)=list(paste("v",1:n,sep=""),paste("f",1:ndim,sep=""))
 A0=A

 # random numbers
 al=rnorm(n, large, sd)
 al[al>1]=1
 as=rnorm(n*ndim-n, small, sd)

 # replace 1's by large and 0's by small
 A[A==1]=al
 A[A==0]=as

 # rotate to pc
 evec=diag(ndim)
 if( pc ){
  tAA=t(A)%*%A
  evec=eigen(tAA)$vectors
  A=A%*%(evec)
  dd=(-1)^!((colSums(A>0) > n/2))
  A=A%*%diag(dd)
  evec=evec%*%diag(dd)
 }
 dimnames(A)=list(paste("v",1:n,sep=""),paste("f",1:ndim,sep=""))

 # reorder
 if( reorder == 1 ){
  resr=reorder_fa( A, sortcols=0 )
  A=resr$loadings
 }
 else if( reorder == 2 ){
  resr=reorder_fa( A, sortcols=1 )
  A=resr$loadings
 }
 else{
  resr=list(row_order=1:n, col_order=1:ndim)
 }
 A0=A0[resr$row_order,resr$col_order]

 return( list(loadings01=A0, loadings=A, rotmat=evec
              , row_order=resr$row_order, col_order=resr$col_order) )

} # end of gendatafa_A
