#' Generate Systematic Missing Data
#'
#' @param n total # of items
#' @param nG total # of forms
#' @param nC # of common items shared by the neighboring forms
#' @param sort = 1 to sort the result
#' @param print = 1 to print the result
#' @return Location of the missing values in a vector
#' @export
#'

gensysmiss <- function( n, nG, nC, sort=0, print=0 ){
 # generate systematic missing pattern
 # Shin-ichi Mayekawa
 # 130127,28
 #
 # Args:
 #
 #  n    total # of items
 #  nG   total # of forms
 #  nC   # of common items shared by the neighboring forms
 #
 #  sort = 1 to sort the result
 #
 #
 # Value
 #  locmiss  a list of length nG
 #           containing the location of missing items for each form
 #
 #
 #  Example of n=18, nG=3, nC=3:    "."  indicates NA
 #
 #     f1 f2 f3
 #  i1  s  .  .
 #  i2  s  .  .
 #  i3  s  .  .
 #  i4  s  .  .
 #  i5  c  c  .   start of the 2nd form is the end of the first - nC + 1
 #  i6  c  c  .
 #  i7  c  c  .
 #  i8  .  s  .
 #  i9  .  s  .
 #  i10 .  s  .
 #  i11 .  s  .
 #  i12 .  c  c  start of the 3rd form is the end of the 2nd - nC + 1
 #  i13 .  c  c
 #  i14 .  c  c
 #  i15 .  .  s
 #  i16 .  .  s
 #  i17 .  .  s
 #  i18 .  .  s
 #
 # Note that the middle forms (f2 above) is longer than
 # the first and the last forms because the middle forms
 # have two sets of common items.
 #

 # distribute items to nG forms
 # total # of items to be shared by the neighboring forms
 ncommon=(nG-1)*nC
 # total # of solo items
 nsolo=n-ncommon

 if( nsolo < 0 ){
  Print("Error: nsolo is negative.")
  Print(n,nG,nC)
  return(NULL)
 }

 # # of solo items in each form: may not be an integer
 nsolo1=nsolo/nG
 if( round(nsolo1) == nsolo1 ) ok=1
 else ok=0
 if( print > 0 ){
  Print(n,nG,nC,"/", ncommon,nsolo, nsolo1, ok
        , fmt="i1 i1 i1 i1 i1 i1 3.1 i1")
 }

 fromto=matrix(NA,nG,2)
 fromto[1,1]=1
 fromto[1,2]=nsolo1+nC
 if( nG > 2 )
  for( k in 2:(nG-1) ){
   fromto[k,1]=fromto[k-1,2]-nC+1
   fromto[k,2]=fromto[k,1]+2*nC+nsolo1-1
  }
 fromto[nG,1]=fromto[nG-1,2]-nC+1
 fromto[nG,2]=n
 nitems=apply(fromto,1,function(x)x[2]-x[1]+1)

 # integer version
 fromto1=floor(fromto)
 nitems1=apply(fromto1,1,function(x)x[2]-x[1]+1)

 if( print > 0 ){
  Print(fromto, nitems, "/", fromto1, nitems1, fmt="4.1 i2")
 }

 if( any(nitems <= 0) ){
  Print("Error:")
  return(NULL)
 }


 G=matrix(1,n,nG)
 for( k in 1:nG ){
  locmiss1=setdiff(1:n,fromto1[k,1]:fromto1[k,2])
  G[locmiss1,k]=NA
 }
 if( sort ){
  od=order(runif(n))
  G=G[od,]
 }
 sumG=apply(G,1,sum, na.rm=1)

 locmiss=NULL
 for( k in 1:nG ){
  locmiss=append(locmiss,list(which(is.na(G[,k]))))
 }


 if( print > 0 ){
  Print(n,nG,nC
        , "/", G, sumG,fmt="i2 i2 i2 i2 i1")
  Print(locmiss)
 }

 return( locmiss )

} # end of gensysmiss



