#' Weighted Orthogonal Procrustes Rotation
#' by Koschat and Swayne's method
#'
#' @param A The matrix to be rotated
#' @param C The target matrix
#' @param D2 The weight matrix
#' @param Tinit initial value of T matrix
#' @param Aainit = Aa matrix which stays constant in this function.
#' @param maxiter max # of iterations
#' @param eps convergence criterion for rmse
#' @param epsd convergence criterion for maximum absolute differences.
#' @param print = 1 to print the result
#'
#' @param SQUAREM = 3 :   See the help of iSQUAREM in lazy.accel package.
#' @param nSQUAREM = 1 :   See the help of iSQUAREM in lazy.accel package.
#' @param minalpha = -999 :   See the help of iSQUAREM in lazy.accel package.
#' @param  maxalpha = -1 :   See the help of iSQUAREM in lazy.accel package.
#' @param  always = 1 :   See the help of iSQUAREM in lazy.accel package.
#' @param  reset1 = 0 :   See the help of iSQUAREM in lazy.accel package.
#' @param  reset2 = 1 :   See the help of iSQUAREM in lazy.accel package.
#'
#' @details
#' This program minimizes \cr
#'    RSS = Tr( ( C  - A T ) D2 ( C - A T )' ) \cr
#'  w.r.t T subject to T'T = TT' = I, \cr
#' using  Koschat and Swayne (1991) algorithm. \cr
#' The missing elements of C matrix will be estimated so that they also
#' minimize RSS.
#'
#' @return
#' A list of B=A T, T, Cm, A, C, D2, rmse, rmse0 \cr
#' where B is the rotated matrix, T is the rotation matrix, \cr
#' Cm is the target matrix with its missing elements replaced by LSE, \cr
#' rmse and rmse0 are, resp, weighted and unweighted RMSE.
#'
#' @examples
#' seed <- 1701; set.seed(seed)
#' nvar <- 20; ndim <- 4
#' errstd <- 0.1
#' resg <- gendataWmm( 1, nvar, ndim )
#' A <- resg$A; Cg <- resg$C; Tg <- resg$T
#' C <- resg$C + errstd*matrix(rnorm(nvar*ndim),nvar)
#' pmiss <- 0.5
#' locmiss <- unique( sample(1:(nvar*ndim), pmiss*(nvar*ndim), replace=1) )
#' C[locmiss] <- NA
#' V <- diag(ndim)
#' V <- matrix(rnorm(nvar*ndim),nvar); V <- t(V)%*%V/nvar
#' Print(A,C,V)
#' res <- procotWKS( A, C, V )
#'
#' @references
#' Koschat, M.A., and Swayne, D.F., (1989). A weighted Procrustes criterion.
#' Psychometrika, 56(2), pp. 229-239.
#'
#' @export
#'

procotWKS <- function( A, C, D2, Tinit=diag(ncol(A)), Aainit=NULL
                       , maxiter=200, eps=1e-6, epsd=1e-3
                       , SQUAREM=3, nSQUAREM=1
                       , minalpha=-999, maxalpha=-1, always=0
                       , reset1=0, reset2=1
                       , print=1 ){
 # Weighted Orthogonal Procrustes Rotation by Koschat and Swayne (1991)
 # PKA
 # Shin-ichi Mayekawa
 # 111225,26
 # print < 0, Tinit: 111227
 # new method for rmse: 120104
 # D2 renamed as D: 120104
 # Xainit, tAA: 120113
 # comments: 120128
 # initial T: 130301
 # comments: 130302
 # D renamed back as D2: 130307
 # 130312
 # bugfix for Xainit: 20150607
 # X->A, Y->C: 20170526ToyamaANA,TamaC
 # missing: 20170528,29
 # lmax -> maxiter: 20170530dnc
 #
 # Args:
 #  A       nv x nr matrix to be rotated
 #  C       nv x nr target matrix
 #  D2       nr x nr weight matrix (known)
 #  Tinit   initial value of T matrix
 #  Aainit  initial value of As matrix
 #  maxiter    max # of iterations
 #  eps     eps for the relative improvement of RMSE
 #  print   to control the print level
 #
 # Returns:
 #  T      nr x nr matrix
 #
 #
 # This program minimizes
 #    RSS = Tr( ( C  - A T ) D2 ( C - A T )' )
 #  w.r.t T subject to T'T = TT' = I.
 #
 #
 # rmse = sqrt(RSS/(nv*nr)) will be reported.
 #
 #

 getrmse <- function( param ){
  Cs1=Cs
  Cs1[1:nv,][locmiss]=param
  rmse=sqrt( sum( (t(Cs1-B)%*%(Cs1-B))*D2 ) / (nv*nr) )
  return( rmse )
 } # getrmse


 # const
 nv=nrow(A); nr=ncol(A)

 # constant: Aa, rho^2, and As
 Aagiven=0
 if( is.null(Aainit) ){
  tAA=t(A)%*%A
  ev=eigen(tAA); eva=matrix(ev$values,length(ev$values),1)
  r2=1.1*max(abs(eva))
  Aa=chol(r2*diag(nr)-tAA)
 }
 else{
  Aa=Aainit
  Aagiven=1
 }
 As=rbind(A,Aa)

 # missing elements of the target
 locmiss=which( is.na(C) )
 locnmiss=which( !is.na(C) )
 nmiss=length(locmiss)
 pmiss=nmiss/nv/nr
 Cm=C
 Cm[locmiss]=A[locmiss]

 # initial values
 if( all( Tinit == diag(nr)) ) Tgiven=0
 else Tgiven=1

 if( !Tgiven ){
  # simple procrustes
  CAAC=t(Cm)%*%A%*%t(A)%*%Cm
  ev=eigen(CAAC); eva=matrix(ev$values,length(ev$values),1); eve=ev$vectors
  rm(ev)
  loc=which(eva > 0); eva=eva[loc]; eve=eve[,loc]
  SQR=eve%*%Diag(sqrt(1/eva))%*%t(eve)
  Tinit=t(A)%*%Cm%*%SQR
 }

 T=Tinit
 Tp=T
 rmsep=sqrt( sum( (t(Cm-A)%*%(Cm-A))*D2 ) / (nv*nr) )
 # same as  sqrt( sum( diag(t(Cm-A)%*%(Cm-A)%*%D2) ) / (nv*nr) )
 converged=0
 B=A

 if (print >= 1){
  cat("\n"); cat("\n"); cat("\n")
  cat("Weighted Orthogonal Procrustes Rotation:\n")
  cat(" minimizing\n")
  cat("    tr(  (C  -  A T) D2 (C  -  A T)' ) \n")
  cat(" w.r.t  orthogonal T \n")
  cat("\n")
  cat("Input parameters:\n")
  cat(" Size of A and C matrices: ", nv, " x ", nr, "\n")
  cat(" Aagiven = ", Aagiven, ",  Tgiven = ", Tgiven, "\n")
  cat(" # of missing elements in C = ", nmiss, "(", pmiss, ")\n")
  cat(" max # of iterations = ", maxiter, " eps for relative rmseimp = "
      , eps, "\n")
  cat(" epsd for the max abs differences of T =", epsd, "\n")
  cat(" SQUAREM=", SQUAREM,",  nSQUAREM=", nSQUAREM
      , " minalpha=", minalpha,",  maxalpha=", maxalpha
      , " reset1=", reset1,",  reset2=", reset2, "\n")
  cat(" print = ", print, "\n")
  cat(" initial rmse =", rmsep, "\n")
  if( print >= 3 ){
   Print(A,C,"/",D2, T,fmt="6.3")
  }
 }


 if( SQUAREM > 0  &&  nmiss > 0 ){
  ########################################################################
  # iSQUAREM
  param=Cm[locmiss]
  iSQUAREM=generate_iSQUAREM( param, print=1 )
  ########################################################################
 }

 # main iteration
 for (llll in 1:maxiter){

  # missing elements of the target
  if( nmiss > 0 ){
   Cm[locmiss]=B[1:nv,][locmiss]
  }


  ###################################################################
  if( nmiss > 0  &&  SQUAREM > 0  &&  nSQUAREM <= llll ){
   param=Cm[locmiss]
   temp=iSQUAREM( param
                  , enforce_constraints=NULL, badness_of_fit=getrmse
                  , SQUAREM=SQUAREM, minalpha=minalpha, maxalpha=maxalpha
                  , bof_value=NULL, always=always, reset1=reset1, reset2=reset2
                  , print=max(0,print-1), debug=0 )
   rmse0=temp$critval
   param=temp$param
   Cm[locmiss]=param
  }
  ###################################################################


  # update orthogonal T
  Cs=rbind(Cm,Aa%*%T)
  udv=svd(D2%*%t(Cs)%*%As)
  T=udv$v%*%t(udv$u)

  # rotated matrix
  B=As%*%T

  # check convergence in terms of As and Cs:  not A and Cm
  rmse=sqrt( sum( (t(Cs-B)%*%(Cs-B))*D2 ) / (nv*nr) )
  rmseimp=rmsep-rmse
  rmseimpr=rmseimp/(rmsep+1e-7)
  maxadt=max(abs(T-Tp))


  if ( print >= 2 )
   Print(llll, rmse, rmseimpr, maxadt, fmt="i3, 7.6 7.6 10.8")

  if ( rmseimpr <= eps  &&  maxadt <= epsd ) {
   converged=1
   break
  }

  # next iteration
  Tp=T; rmsep=rmse

 } # end of llll loop

 iter=llll


 if ( !converged  && print >= 0 )
  cat("\n Warning(procotWKS): Iteration did not converge.\n\n")

 B=A%*%T
 rmse=sqrt( sum( (t(Cm-B)%*%(Cm-B))*D2 ) / (nv*nr) )
 rmse0=sqrt( sum( (t(Cm-B)%*%(Cm-B)) ) / (nv*nr) )

 if ( print > 0 ){
  # Print(llll, rmsep, rmse, rmseimp, rmseimpr, maxadt
  #      , fmt="i3, 7.5 7.5 7.5 10.8")
  cat("\n\nWeighted Procrustes Rotation\n")
  cat(" minimizing\n")
  cat("    tr(  (C  -  A T) D2 (C  -  A T)' ) \n")
  cat(" w.r.t  orthogonal T \n")
  cat(" Size of A and C matrices: ", nv, " x ", nr, "\n")
  cat(" Aagiven = ", Aagiven, ",  Tgiven = ", Tgiven, "\n")
  cat(" # of missing elements in C = ", nmiss, "(", pmiss, ")\n")
  cat(" SQUAREM=", SQUAREM,",  nSQUAREM=", nSQUAREM
      , " minalpha=", minalpha,",  maxalpha=", maxalpha
      , " reset1=", reset1,",  reset2=", reset2, "\n")
  cat(" # of iterations required for convergence =", iter
      , " with eps=", eps, "  and epsd=", epsd,"\n")
  cat("\n Final Weighted RMSE = ", rmse, "  and Unweighted RMSE =", rmse0,"\n")
  cat("\n")
  Print(T, D2, fmt="7.3")
 }

 # returning the result
 res=named_list( B, T, A, C, D2, Cm, rmse, rmse0 )
 return( res )


} # end of procotWKS




