#' Weighted Expanded Orthogonal Procrustes Rotation \cr
#' The Ultimate Model
#'
#' @param A The input matrix to be rotated
#' @param C The target matrix
#' @param V The weight matrix (known)
#' @param estm = 1 to estimate m
#' @param estT = 0 to avoid estimation of T
#' @param estW = 1 to estimate scalar W \cr
#' = 2 to estimate diagonal W matrix
#' @param estW2 = 1 to estimate diagonal W2 matrix
#' @param W Initial value of W or NULL
#' @param T Initial value of T or NULL
#' @param m Initial value of m or NULL
#' @param W2 Initial value of W2 or NULL
#' @param maxiter max # of iterations
#' @param eps convergence criterion for rmse
#' @param maxiter2 max # of iterations for procotWKS
#' @param eps2 convergence criterion for procotWKS
#' @param print = 1 to print result
#'
#' @return A list of: \cr
#' B=A \%*\% W \%*\% T + matrix(1,nv,1)\%*\%t(m), T, m, W, W2
#' , Cm the updated target matrix, rmse=sqrt(RSS)
#'
#' @details
#' This program finds those A, m, W, W2  and orthonormal T which minimize \cr
#'   RSS = tr( ( C  - Chat )  V ( C - Chat )' ) \cr
#'   where Chat = A \%*\% W \%*\% T \%*\% W2 + matrix(1,nv) \%*\% t(m),
#'   W  and W2 are diagonal matrices, T is an orthonormal rotation matrix, \cr
#'   m is the column mean vector. \cr
#'\cr
#' Note that when estW=0, estW2=0 and estm=0, and the target matrix does not
#' have missing elements, \cr
#' this is equivalent to the usual orthogonal Procrustes rotation.
#'
#' @examples
#'  seed <- 1701; n <- 20; nr <- 2; pmiss <- 0; errstd <- 0.05
#'  set.seed(seed)
#'  resg <- gendataWmm( 1, n, nr, genm=1, genW=2, genW2=1, genV=1
#'                    , pmiss=pmiss, errstd=errstd )
#'  A <- resg$A; V <- resg$V
#'  C <- resg$C; Tg <- resg$T; mg <- resg$m; Wg <- resg$W; W2g <- resg$W2
#'  res <- procotWeW( A, C, V, estm=1, estW=2, estW2=1, print=1 )
#'  T <- res$T; m <- res$m; W <- res$W; W2 <- res$W2
#'  Print( Tg-T, mg-m, fmt="7.3" )
#'  Print( Wg-W, W2g-W2, fmt="7.3" )
#'
#' @export
#'

procotWeW <- function( A, C, V, estm=0, estT=1, estW=0, estW2=0
                 , W=NULL, T=NULL, m=NULL, W2=NULL
                 , maxiter=200, eps=1e-6, maxiter2=20, eps2=1e-6, print=2 ){
 # Weighted Expanded Orthogonal Procrustes Rotation: The Ultimate Model
 # Shin-ichi Mayekawa
 # 111220,22
 # do not normalize Ds when estW=0: 111223
 # Cm -> C, etc :111226
 # procoteW3.R renamed as EOPR.R: 111226
 # The Ultimate Model: 111227
 # weighted flag: 111228
 # no D1 or D2: 111228
 # weighted with V: 120103.04
 # V2 renamed as V: 120104
 # bugfix for reflection of W and W2: 120105
 # WEOPR renamed as procotWeW: 20150607
 # procotWKS returns a list: 20170526Tamac
 # lmax -> maxiter: 20170530dnc
 #
 # Args:
 #
 #  A       nv x nr matrix to be rotated
 #  C       nv x nr target matrix
 #
 #  V       nR x nr weight matrix (inverse of the dispersion matrix)
 #
 #  estm    = 1 to estimate m
 #  estT    = 1 to estimate orthogonal T
 #  estW    = 1 to estimate scalar   W
 #          = 2 to estimate diagonal W
 #  estW2   = 1 to estimate diagonal W2
 #
 #  maxiter    max # of iterations
 #  eps     eps for the relative improvement of RMSE
 #  maxiter2 max # of iterations for the Weighted Orthogonal Procrustes Rotation
 #  eps2   eps for the relative improvement of RMSE of Weighted Procrustes
 #
 #  print   to control the print level
 #
 #  m, W, T, and W2   initial values
 #
 #
 # Returns:
 #  list(B=B, T=T, m=m, W=W, W=W2, C=C, rmse=rmse)
 #  where B is the result of the rotation,
 #  C is the nv x nr matrix with the missing elements estimated by LS.
 #
 # This program minimizes
 #    RSS = tr( ( C  - (A W T W2 + 1*m') ) V ( C - (A W T W2 + 1*m') )' )
 #  w.r.t  T, m, W and W2 subject to T'T = TT' = I  and  W, W2 = diag.
 #
 # When both W and W2 are estimated, W2 will be nornalized so that
 #   sum(W2*W2) = 1
 #
 # The missing elements of C matrix will be estimated so that they also
 # minimize RSS.
 #
 # rmse = sqrt(RSS/(nv*nr)) will be reported.
 #
 #
 # Needs:
 #  procotWKS.R
 #

 getrmse <- function( title, A, C, m, W, T, W2, D1, D2, V, nv, nr, rmsep ){
  B=A%*%W%*%T%*%W2+matrix(1,nv,1)%*%t(m);
  DCD=C;
  # rmse=sqrt( sum(((DCD-B)%*%V)^2) / (nv*nr) );
  rmse=sqrt( sum( (t(DCD-B)%*%(DCD-B))*V ) / (nv*nr) );
  cat(title, ":", rmse, "\n")
  if( rmsep < rmse ){
    cat(title, ": **** failure !! ****", rmsep - rmse, " \n")
  }
  return( rmse )
 } # end of getrmse



 # const
 nv=nrow(A); nr=ncol(A);

 # center A matrix when m is estimated
 Aorg=A
 if (estm == 1){
  amean=as.matrix(apply(A,2,mean),ncol(A),1);
  A=A-matrix(1,nv,1)%*%t(amean);
 }
 traceAtA=tr(t(A)%*%A);


 # missing target elements
 locmiss=is.na(C); # logical matrix of the same size as C;
 nmiss=sum(locmiss);
 pmiss=nmiss/(nv*nr);
 if (nmiss > 0)
  C[locmiss]=Aorg[locmiss]


 # initial values
 if( is.null(W) ) W=diag(nr);
 if( is.null(m) ) m=matrix(0,nr,1);
 if( is.null(T) ) T=diag(nr);
 if( is.null(W2) ) W2=diag(nr)


 DCD=C;
 rmsep=sqrt( sum( (t(DCD-A)%*%(DCD-A))*V ) / (nv*nr) )
 rmse0=rmsep;
 converged=0;

 # weighted Procrustes or not
 if( V == diag(nr)  &&  estW2 ==0  && W2 == diag(nr) ) weighted=0
 else weighted=1

  if (print >= 2){
  cat("\n"); cat("\n"); cat("\n");
  cat("Weighted Expanded Orthogonal Procrustes Rotation")
  cat("  (The Ultimate Model):\n")
  cat(" Minimizing ")
  cat(" RSS = tr( ( C  - (A W T W2 + 1*m') ) V ( C - (A W T W2 + 1*m') )' )\n")
  cat("  w.r.t T, m, W, and W2 \n")
  cat("  subject to T'T = TT' = I  and W, W2 = diagonal \n")
  cat(" The missing elements of the targetmatrix C will be updated.\n")
  cat("\n")
  cat("Input parameters:\n")
  cat(" Size of A and C matrices: ", nv, " x ", nr, "\n")
  cat(" estimation of m = ", estm, "\n")
  cat(" estimation of T = ", estT, " with weighted = ", weighted, "\n")
  cat(" estimation of W = ", estW, ", estimation of W2 = ", estW2, "\n")
  cat(" max # of iterations = ", maxiter
    , ", eps for relative rmseimp = ", eps, "\n")
  cat(" print = ", print, "\n")
  cat(" # of missing elements in C = ", nmiss, "(", pmiss, ")\n\n")
 }

 Tp=T; Wp=W; W2p=W2


 if( print >= 4 )
   rmse0=getrmse( "initial", A, C, m, W, T, W2, D1, D2, V, nv, nr, 9999 )


 # main iteration
 for (llll in 1:maxiter){


  # update orthogonal T;
  if( estT == 1 ){
   if( weighted == 1 ){
    # weighted Orthogonal Procrustes by  Koschat and Swayne
    T=procotWKS( A%*%W, DCD%*%solve(W2), W2%*%V%*%W2, Tinit=T
               , print=-1, eps=eps2, maxiter=maxiter2 )$T
   } else {
    # simple Orthogonal Procrustes
    CAAC=t(DCD)%*%A%*%W%*%W%*%t(A)%*%DCD;
    ev=eigen(CAAC); eva=matrix(ev$values,length(ev$values),1); eve=ev$vectors
    loc=which(eva > 0); eva=eva[loc]; eve=eve[,loc];
    SQR=eve%*%Diag(sqrt(1/eva))%*%t(eve);
    T=t(A%*%W)%*%DCD%*%SQR;
   }
   if( print >= 4 )
      rmse0=getrmse( "after T", A, C, m, W, T, W2, D1, D2, V, nv, nr, rmse0 )
  }

  # update m
  if (estm == 1) {
   m=as.matrix(apply(DCD-A%*%W%*%T%*%W2,2,mean),nr,1);
    if( print >= 4 )
      rmse0=getrmse( "after m", A, C, m, W, T, W2, D1, D2, V, nv, nr, rmse0 )
  }

  # update W
  if (estW == 1 ) {
   d=tr(t(A%*%T%*%W2)%*%DCD)/traceAtA;
   # make d positive;
   if ( d < 0 ){
    d=-d; T=-T;
   }
   W=diag(rep(d,nr))
  }
  else if( estW == 2 ){
   W=solve((T%*%W2%*%V%*%W2%*%t(T))*(t(A)%*%A))%*%
     vecdiag(t(A)%*%DCD%*%V%*%W2%*%t(T))
   W=Diag(W)
   if( print >= 4 )
      rmse0=getrmse( "after W", A, C, m, W, T, W2, D1, D2, V, nv, nr, rmse0 )
  }

  # update W2
  if( estW2 >= 1 ){

   W2=solve((t(T)%*%W%*%t(A)%*%A%*%W%*%T)*V)%*%
      vecdiag(t(T)%*%W%*%t(A)%*%DCD%*%V)
   W2=Diag(W2)

  # W20=solve(Diag(t(T)%*%W%*%t(A)%*%A%*%W%*%T))%*%Diag(t(T)%*%W%*%t(A)%*%DCD)
  # Print(W20-W2)

   if( print >= 4 )
      rmse0=getrmse( "after W2", A, C, m, W, T, W2, D1, D2, V, nv, nr, rmse0 )
  }

  # normalize W2
  if( estW >= 1  && estW2 >= 1 ){
   dd=sqrt(sum(W2*W2))
   W2=W2/dd
   W=W*dd
  }


  # rotated matrix
  B=A%*%W%*%T%*%W2 + matrix(1,nv,1)%*%t(m);

#printm(A,W,T,m)
#printm(C,B)


  # check convergence
  B=A%*%W%*%T%*%W2 + matrix(1,nv,1)%*%t(m);
  rmse=sqrt( sum( (t(DCD-B)%*%(DCD-B))*V ) / (nv*nr) )
  rmseimp=rmsep-rmse;
  rmseimpr=rmseimp/(rmsep+1e-7);
  maxadt=max(abs(T-Tp));
  maxadW=max(abs(W-Wp));
  maxadW2=max(abs(W2-W2p));

  if ( print >= 2 )
   Print(llll, rmsep, rmse, rmseimp, rmseimpr, maxadt, maxadW, maxadW2
         , sep="  ", digits=3);

  if ( rmseimpr < eps || maxadt < eps) {
   converged=1;
   break;
  }

  # update the target;
  if ( nmiss > 0 ){
   C[locmiss]=B[locmiss];
   DCD=C;
  }

  # next iteration;
  Tp=T; Wp=W; W2p=W2
  rmsep=rmse;


 } # end of llll loop;


 if ( !converged  &&  print >= 0 )
  cat("\n Warning(procote): Iteration did not converge.\n\n")

 # update m using the original A
 if (estm == 1) {
  m=as.matrix(apply(DCD-Aorg%*%W%*%T%*%W2,2,mean),nr,1);
 }

 # reflect W2
 if( any( vecdiag(W2) < 0 ) ){
   locneg=which(vecdiag(W2) < 0)
   dd=rep(1,nr); dd[locneg]=-1; dd=diag(dd)
   W2=dd%*%W2
   T=T%*%dd
 }

 # reflect W
 if( any( vecdiag(W) < 0 ) ){
   locneg=which(vecdiag(W) < 0)
   dd=rep(1,nr); dd[locneg]=-1; dd=diag(dd)
   W=W%*%dd
   T=dd%*%T
 }

 B=Aorg%*%W%*%T%*%W2+matrix(1,nv,1)%*%t(m);
 DCD=C;
 dif=DCD-B;
 rmse=sqrt( sum( (t(DCD-B)%*%(DCD-B))*V ) / (nv*nr) )
 if ( print > 0 ){
  if ( nmiss > 0 ){
   cat("Iteration Terminated.\n")
  Print(llll, rmsep, rmse, rmseimp, rmseimpr, maxadt, sep="  ", digits=5);
  }
  cat("\n\nWeighted Expanded Orthogonal Procrustes Rotation")
  cat("  (The Ultimate Model)\n")
  cat(" Minimizing ")
  cat(" RSS = tr( ( C  - (A W T W2 + 1*m') ) V ( C - (A W T W2 + 1*m') )' )\n")
  cat("  w.r.t T, m, W, and W2 \n")
  cat("  subject to T'T = TT' = I  and  W, W2 = diagonal \n")
  cat(" Size of A and C matrices: ", nv, " x ", nr, "\n")
  cat(" estimation of m = ", estm, "\n")
  cat(" estimation of T = ", estT, " with weighted = ", weighted, "\n")
  cat(" estimation of W = ", estW, ", estimation of W2 = ", estW2, "\n")
  cat(" max # of iterations = ", maxiter, ", eps for relative rmseimp = "
      , eps, "\n")
  cat(" print = ", print, "\n")
  cat(" # of missing elements in C = ", nmiss, "(", pmiss, ")\n")
  cat(" Final RMSE = ", rmse, "\n")
  cat("\n");
  Print(T, m, W, W2, V, digits=3);
 };

 # returning the result
 res=list(B=B, T=T, m=m, W=W, W2=W2, C=C, rmse=rmse);
 return( res );


} # end of procotWeW


