#' Expanded Orthogonal Procrustes Rotation
#'
#' @param A The input matrix to be rotated
#' @param C The target matrix
#' @param estm = 1 to estimate m
#' @param estW = 1 to estimate scalar W \cr
#' = 2 to estimate diagonal W
#' @param minW minimum value of the elements of W
#' @param T Initial value of T or NULL
#' @param W Initial value of W or NULL
#' @param m Initial value of m or NULL
#' @param maxiter max # of iterations
#' @param eps convergence criterion for rmse
#' @param print = 1 to print result
#'
#' @return A list of: \cr
#' B=A \%*\% W \%*\% T + matrix(1,nv,1)\%*\%t(m), T, W,
#' , C the updated target matrix, rmse=sqrt(RSS)
#'
#' @details
#' This program finds those A, m, W  and orthonormal T which minimize \cr
#'   RSS = tr( ( C  - Chat )' ( C - Chat ) ) \cr
#'   where Chat = A \%*\% W \%*\% T + matrix(1,nv) \%*\% t(m),
#'   W is a diagonal Matrix, T is an orthonormal rotation matrix, \cr
#'   m is the column mean vector. \cr
#'\cr
#' Note that when estW=0 and estm=0, and the target matrix does not
#' have missing elements, \cr
#' this is equivalent to the usual orthogonal Procrustes rotation.
#'
#' @examples
#'  seed <- 1701; n <- 20; nr <- 2; pmiss <- 0; errstd <- 0.05
#'  set.seed(seed)
#'  resg <- gendataWmm( 1, n, nr, genm=1, genW=2, pmiss=pmiss, errstd=errstd)
#'  A <- resg$A
#'  C <- resg$C; Tg <- resg$T; mg <- resg$m; Wg <- resg$W
#'  res <- procoteW( A, C, estm=1, estW=2, print=1 )
#'  T <- res$T; m <- res$m; W <- res$W
#'  Print( Tg-T, mg-m, Wg-W, fmt="7.3")
#'
#'
#' @export
#'

procoteW <- function( A, C, estm=0, estW=0, minW=1e-3
                    , T=NULL, W=NULL, m=NULL
                    , maxiter=200, eps=1e-6, print=2 ){
 # Expanded Orthogonal Procrustes Rotation
 # Shin-ichi Mayekawa
 # 111220,21,22
 # do not normalize Ds when estW=0: 111223
 # Cm -> C, etc :111227
 # print < 0, return llll : 120109
 # remove D1 and D2: 120115
 # initial value: 120115
 # minW: 130124
 # lmax -> maxiter: 20170530dnc
 #
 # Args:
 #  A       nv x nr matrix to be rotated
 #  C       nv x nr target matrix
 #  estm    = 1 to estimate m
 #  estW    = 1 to estimate scalar W
 #          = 2 to estimate diagonal W
 #  minW    minimum value of W matrix
 #  maxiter    max # of iterations when there are NAs in C
 #  eps     eps for the relative improvement of RMSE
 #  print   to control the print level
 #
 #  m, W, T  initial values
 #
 # Returns:
 #  list(B=B, T=T, m=m, W=W, C=C, rmse=rmse)
 #  where B is the result of the rotation,
 #  C is the nv x nr matrix with the missing elements estimated by LS.
 #
 # This program minimizes
 #    RSS = tr( ( C  - (A W T + 1*m') )' ( C - (A W T + 1*m') ) )
 #  w.r.t  T, m and W subject to T'T = TT' = I  and D1, D2, W = diag.
 #
 #
 # The missing elements of C matrix will be estimated so that they also
 # minimize RSS.
 #
 # rmse = sqrt(RSS/(nv*nr)) will be reported.
 #
 #
 #
 #
 # When all the elements of a column of C is zero. ###############
 #
 #
 #

 getrmse <- function( title, A, C, m, W, T, nv, nr, rmsep ){
  B=A%*%W%*%T+matrix(1,nv,1)%*%t(m);
  rmse=sqrt( sum((C-B)^2) / (nv*nr) );
  cat(title, ":", rmse, "\n")
  if( rmsep < rmse ){
    cat(" ",title, ": **** failure !! ****", rmsep - rmse, " \n")
  }
  return( rmse )
 }


 # const
 nv=nrow(A); nr=ncol(A);

 # column center A if m is present
 Aorg=A
 if (estm == 1){
  amean=as.matrix(apply(A,2,mean),ncol(A),1);
  A=A-matrix(1,nv,1)%*%t(amean);
 }
 traceAtA=tr(t(A)%*%A);

 # initial value
 mgiven=1;
 if( is.null(m) ){
  m=matrix(0,nr,1);
  mgiven=0
 }
 Tgiven=1;
 if( is.null(T) ){
  T=diag(nr);
  Tgiven=0
 }
 Wgiven=1;
 if( is.null(W) ){
  W=diag(nr);
  Wgiven=0
 }

 # missing target elements
 locmiss=is.na(C); # logical matrix of the same size as C;
 nmiss=sum(locmiss);
 pmiss=nmiss/(nv*nr);
 if (nmiss > 0)
  C[locmiss]=Aorg[locmiss]


 if (print >= 2){
  cat("\n"); cat("\n"); cat("\n");
  cat("Expanded Orthogonal Procrustes Rotation:\n")
  cat(" minimizing  ssq(  C  - (A W T + 1*m') )\n")
  cat("  w.r.t  T, m and W \n")
  cat("  subject to T'T = TT' = I  and D1, D2, W = diagonal \n")
  cat(" The missing elements of the targetmatrix C will be updated.\n")
  cat("\n")
  cat("Input parameters:\n")
  cat(" Size of A and C matrices: ", nv, " x ", nr, "\n")
  cat(" estimation of m = ", estm, ", estimation of W = ", estW, "\n")
  cat(" minimum value for W =", minW, "\n")
  cat(" max # of iterations = ", maxiter
    , ", eps for relative rmseimp = ", eps, "\n")
  cat(" mgiven =", mgiven, ", Wgiven =", Wgiven, ", Tgiven =", Tgiven,"\n")
  cat(" print = ", print, "\n")
  cat(" # of missing elements in C = ", nmiss, "(", pmiss, ")\n\n")
 }


 # initial values
 Tp=T;
 rmsep=sqrt(sum((C-A%*%W%*%T)^2)/(nv*nr));
 rmse0=rmsep;
 converged=0;


 # main iteration
 for (llll in 1:maxiter){


  # update orthogonal T;
  CAAC=t(C)%*%A%*%W%*%W%*%t(A)%*%C;
  ev=eigen(CAAC); eva=matrix(ev$values,length(ev$values),1); eve=ev$vectors;
  loc=which(eva > 0); eva=eva[loc]; eve=eve[,loc];
  if( length(eva) < nr ){
   Print(C,A,W,ev$values)
  }
  eva[eva<0]=0
  SQR=eve%*%Diag(sqrt(1/eva))%*%t(eve);
  T=t(A%*%W)%*%C%*%SQR;
#  rm(ev,CAAC,SQR);

  if( print >= 4 )
   rmse0=getrmse( "procoteW: after T", A, C, m, W, T, nv, nr, rmsep )

  # update m
  if (estm == 1) {
   m=as.matrix(apply(C-A%*%W%*%T,2,mean),nr,1);
  if( print >= 4 )
   rmse0=getrmse( "procoteW: after m", A, C, m, W, T, nv, nr, rmse0 )
  }

  # update W
  if (estW == 1 ){
   d=tr(t(A%*%T)%*%C)/traceAtA;
   # make d positive;
   if ( d < 0 ){
    d=-d; T=-T;
   }
   W=diag(rep(d,nr))
  }
  else if( estW == 2 ){
   W=solve( Diag(t(A)%*%A) )%*%Diag( t(A)%*%C%*%t(T) )
   diag(W)[diag(W) < minW]=minW
  if( print >= 4 )
   rmse0=getrmse( "procoteW: after W", A, C, m, W, T, nv, nr, rmse0 )
  }

  # rotated matrix
  B=A%*%W%*%T + matrix(1,nv,1)%*%t(m);

#printm(A,W,T,m)
#printm(C,B)


  # check convergence
  rss=sum((C-B)^2)
  rmse=sqrt( rss/nv/nr )
  rmseimp=rmsep-rmse;
  rmseimpr=rmseimp/(rmsep+1e-7);
  maxadt=max(abs(T-Tp));

  if ( print >= 2 )
   Print(llll, rmsep, rmse, rmseimp, rmseimpr, maxadt, fmt="i3 10.6");

  if ( rmseimpr < eps || maxadt < eps) {
   converged=1;
   break;
  }

  # update the target;
  if ( nmiss > 0 ){
   C[locmiss]=B[locmiss];
  }

  # next iteration;
  Tp=T; rmsep=rmse;


 } # end of llll loop;



 if ( !converged && print > 0 )
  cat("\n Warning(procoteW): Iteration did not converge.\n\n")




 # update m using the original A
 if (estm == 1) {
  m=as.matrix(apply(C-Aorg%*%W%*%T,2,mean),nr,1);
 }

 B=Aorg%*%W%*%T+matrix(1,nv,1)%*%t(m);
 dif=C-B;
 rmse=sqrt( sum((C-B)^2) / (nv*nr) );

 if ( print > 0 ){
  if ( nmiss > 0 ){
   cat("Iteration Terminated.\n")
  Print(llll, rmsep, rmse, rmseimp, rmseimpr, maxadt, fmt="i3 10.6");
  }
  cat("\n\nExpanded Orthogonal Procrustes Rotation:\n")
  cat(" minimizing  ssq(  C  - (A W T + 1*m') )\n")
  cat("  w.r.t  T, m and W \n")
  cat("  subject to T'T = TT' = I  and  W = diagonal \n")
  cat(" Size of A and C matrices: ", nv, " x ", nr, "\n")
  cat(" minimum value for W =", minW, "\n")
  cat(" estimation of m = ", estm, ", estimation of W = ", estW, "\n")
  cat(" max # of iterations = ", maxiter, " eps for relative rmseimp = ", eps
  , "\n")
  cat(" print = ", print, "\n")
  cat(" mgiven =", mgiven, ", Wgiven =", Wgiven, ", Tgiven =", Tgiven,"\n")
  cat(" # of missing elements in C = ", nmiss, "(", pmiss, ")\n")
  cat(" Final RMSE = ", rmse, "\n")
  cat("\n");
  Print(T, m, W, fmt="7.4");
 };

 # returning the result
 res=list(B=B, T=T, m=m, W=W, C=C, rmse=rmse, llll=llll);
 return( res );

} # end of procoteW



