#' Procrustes Rotation with Prescribed Factor Correlations
#'
#' @param A The matrix to be rotated
#' @param C The target matrix
#' @param Phia The factor correlation matrix associated with A
#' @param Phib The factor correlation of the rotated factors
#' @param Tinit initial value of T matrix
#' @param maxiter max # of iterations
#' @param eps convergence criterion for rmse
#' @param epsd convergence criterion for maximum absolute differences of Q.
#' @param maxiter2 max # of iterations of procotWKS
#' @param eps2 convergence criterion for rmse of procotWKS
#' @param epsd2 convergence criterion for maximum absolute differences
#'  of procotWKS
#' @param print = 1 to print the result
#'
#' @param SQUAREM = 3 :   See the help of iSQUAREM in lazy.accel package.
#' @param nSQUAREM = 1 :   See the help of iSQUAREM in lazy.accel package.
#' @param minalpha = -999 :   See the help of iSQUAREM in lazy.accel package.
#' @param  maxalpha = -1 :   See the help of iSQUAREM in lazy.accel package.
#' @param  always = 1 :   See the help of iSQUAREM in lazy.accel package.
#' @param  reset1 = 0 :   See the help of iSQUAREM in lazy.accel package.
#' @param  reset2 = 1 :   See the help of iSQUAREM in lazy.accel package.
#'
#'
#' @details
#'  This function finds the factor rotation matrix T of the form: \cr
#'    g=T'f and B=A inv(T') \cr
#'  which minimizes the least squares criterion: \cr
#'   RSS = tr(  (C - B)'(C - B) ) \cr
#'  subject to corr(g)=Phib. \cr
#'  The rotation matrix T can be defined as: \cr
#'     where T = P inv(K) Q D R'  and  QQ'=Q'Q=I, \cr
#'     where corr(f)=Phia=P K2 P'  and  corr(g)=Phib=R D2 R'. \cr
#'  The missing elements of C matrix will be estimated so that they also
#' minimize RSS.
#'
#' @return
#' A list of  B, T, Q, Cm, A, C, Phia, Phib, rmse, \cr
#' where B is the rotated matrix, T is the rotation matrix, \cr
#' Q is the orthogonal matrix which defines T, and \cr
#' Cm is the target matrix with its missing elements replaced by LSE.
#'
#' @examples
#'
#' # Independent Cluster
#' seed <- 1701
#' set.seed(seed)
#'
#' nvar <- 20
#' ndim0 <- 3
#' ps <- 0.1
#' df <- 500
#' phi <- 0.3
#' big=0.8
#'
#' Lambda0 <- gendatafa_A( nvar, ndim0, large=big,small=1-big
#'                   , pc=0, sd=0 )$loadings
#' colnames(Lambda0) <- paste("f",1:ndim0,sep="")
#'
#' Phi <- (1-phi)*diag(ndim0)+phi*matrix(1,ndim0,ndim0)
#' Sigma <- Lambda0%*%Phi%*%t(Lambda0)+ps*diag(nvar)
#' dS <- sqrt(diag(Sigma))
#' Sigma <- diag(1/dS)%*%Sigma%*%diag(1/dS)
#'
#' S <- rWishart( 1, df, Sigma )
#' S <- S[,,1]/df
#' dS <- sqrt(diag(S))
#' S <- diag(1/dS)%*%S%*%diag(1/dS)
#'
#' ndim <- ndim0
#'
#'
#' temp <- lazy.fa::fa_hs( S, ndim=ndim, c="smc" )
#' Lambda <- temp$Lambda
#' psi <- temp$psic
#'
#'
#' Lambda00 <- Lambda0
#' Lambda00[Lambda00==big] <- 1
#' Lambda00[Lambda00==1-big] <- 0
#' Lambda000=Lambda00
#' Lambda000[Lambda00==1]=NA
#'
#' phia <- 0; Phia <- (1-phia)*diag(ndim)+phia*matrix(1,ndim,ndim)
#' table <- NULL
#' for( p in seq(-0.45, 0.45, 0.05) ){
#'  Print(p)
#'  Phib <- (1-p)*diag(ndim)+p*matrix(1,ndim,ndim)
#'  res <- procrustes( Lambda, Lambda0, Phia=Phia, Phib, print=1, maxiter2=500 )
#'  table <- rbind(table,c(p,res$rmse))
#'  Print(res$B)
#' }
#' best=table[,2]==min(table[,2])
#' Print(table,best)
#'
#' @export
#'

procrustes <- function( A, C, Phia=diag(ncol(A)), Phib=NULL, Tinit=NULL
                      , maxiter=500, eps=1e-6, epsd=1e-6
                      , maxiter2=200, eps2=1e-6, epsd2=1e-3
                      , SQUAREM=3, nSQUAREM=1
                      , minalpha=-999, maxalpha=-1, always=0
                      , reset1=0, reset2=1
                      , print=1 ){
 # Procrustes rotation with  prescribed factor correlation matrix
 # Shin-ichi Mayekawa
 # Originally, procotfc: 130301,02
 # Newly written: 20170526ToyamaANAH,Kagayaki
 # 20170528,29
 # lmax -> maxiter: 20170530dnc
 #

 #
 #  disp(f) = Phi_f = Phia = P K^2 P'
 #  disp(g) = Phi_g = Phib = R D^2 R'

 #  g = T' f
 #  B = A inv(T')
 #
 #  T = P inv(K) Q D R'
 #       inv(T)  = R inv(D) Q' K P'
 #       inv(T') = P K Q inv(D) R'
 #
 #      Q = P' K T R inv(D)
 #
 #  rss = trssq( C-A inv(T') )
 #      = trssq( C-A PKQinv(D)R' )
 #      = tr( (CRD-APK Q) D^{-2} (CRD-APK Q)' )
 #


 # crit for SQUAREM
 getrmse <- function( param ){
  Cm1=Cm
  Cm1[locmiss]=param
  rmse=sqrt( sum( (Cm1-B)^2 ) / (nvar*ndim) )
  return( rmse )
 } # getrmse



 # const
 nvar=nrow(A)
 ndim=ncol(A)

 # Decomposition pf Phis
 P=diag(ndim); K=P; R=P; D=P
 if( is.null(Phia) ) Phia=diag(ndim)
 if( is.null(Phib) ) Phib=diag(ndim)
 if( any(Phia != diag(ndim)) ){
  temp=svd(Phia); P=temp$u; K=diag(sqrt(temp$d))
 }
 if( any(Phib != diag(ndim)) ){
  temp=svd(Phib); R=temp$u; D=diag(sqrt(temp$d))
 }

 invD2=solve(D%*%D)
 PK=P%*%K
 APK=A%*%PK
 RD=R%*%D
 invDtR=solve(RD)

 # to be handed to procotWKS
 tAA=t(APK)%*%APK
 ev=eigen(tAA); eva=matrix(ev$values,length(ev$values),1)
 r2=1.1*max(abs(eva))
 Aa=chol(r2*diag(ndim)-tAA)

 # missing elements of the target
 locmiss=which( is.na(C) )
 locnmiss=which( !is.na(C) )
 nmiss=length(locmiss)
 pmiss=nmiss/nvar/ndim
 Cm=C

 # initial
 B=A
 Cm[locmiss]=A[locmiss]
 #      Q = P' K T R inv(D)
 T=procote( A, Cm, print=0 )$T
 Q=t(P)%*%K%*%T%*%R%*%solve(D)

 if ( print > 0 ){
  cat("\n\nProcrustes Rotation with a Prescribed Factor Correlations\n")
  cat("\n This function finds the factor rotation matrix T of the form:")
  cat("\n   g=T'f and B=A inv(T')")
  cat("\n which minimizes the least squares criterion:")
  cat("\n  tr(  (C - B)' (C - B) )")
  cat("\n subject to corr(g)=Phib.")
  cat("\n The rotation matrix can be defined as:")
  cat("\n    where T = P inv(K) Q D R'")
  cat("\n    where corr(f)=Phia=P K2 P'  and  corr(g)=Phib=R D2 R'. \n ")
  cat(" Size of A and C matrices: ", nvar, " x ", ndim, "\n")
  cat(" # of missing elements in C = ", nmiss, "(", pmiss, ")\n")
  cat(" max # of iterations = ", maxiter
      , " eps for relative rmseimp = ", eps, "\n")
  cat(" epsd for the max abs differences of Q =", epsd, "\n\n")
  cat(" SQUAREM=", SQUAREM,",  nSQUAREM=", nSQUAREM
      , " minalpha=", minalpha,",  maxalpha=", maxalpha
      , " reset1=", reset1,",  reset2=", reset2, "\n")
  cat(" The factor correlation matrices of f and g, resp. \n")
  Print(Phia, Phib, fmt="7.3")
  if( print >= 3 ){
   Print(C)
  }
 }



 if( SQUAREM > 0  &&  nmiss > 0 ){
  ########################################################################
  # iSQUAREM
  param=Cm[locmiss]
  iSQUAREM=generate_iSQUAREM( param, print=1 )
  ########################################################################
 }


 # previous values
 Cp=Cm
 Qp=Q
 rmsep=999999


 for( llll in 1:maxiter ){


  # missing elements of the target
  if( nmiss > 0 ){
   Cm[locmiss]=B[locmiss]
  }


  ###################################################################
  if( nmiss > 0 && SQUAREM > 0  &&  nSQUAREM <= llll ){
   param=Cm[locmiss]
   temp=iSQUAREM( param
                  , enforce_constraints=NULL, badness_of_fit=getrmse
                  , SQUAREM=SQUAREM, minalpha=minalpha, maxalpha=maxalpha
                  , bof_value=NULL, always=always, reset1=reset1, reset2=reset2
                  , print=max(0,print-1), debug=0 )
   rmse0=temp$critval
   param=temp$param
   Cm[locmiss]=param
  }
  ###################################################################

  # weighted orthogonal procrustes
  CRD=Cm%*%RD
  Q=procotWKS( APK, CRD, invD2, Tinit=Qp, Aa=Aa
               , maxiter=maxiter2, eps=eps2, epsd=epsd2
               , print=max(0,print-9) )$T
  invtT=PK%*%Q%*%invDtR
  B=A%*%invtT


  # convergence
  rmse=sqrt( sum((Cm-B)^2)/nvar/ndim )
  rmseimp=rmsep-rmse
  rmseimpr=rmseimp/rmsep
  maxadQ=max(abs(Q-Qp))
  if( print >= 2 ){
   Print( llll,rmse, rmseimpr, maxadQ )
   #      , fmt="i3, 7.5 7.5 10.8")
  }
  if( nmiss == 0  || (rmseimpr <= eps  &&  maxadQ <= epsd) ){
   break
  }

  # next iteration
  rmsep=rmse
  Qp=Q


 } # end of llll loop


 iter=llll
 T=t(solve(invtT))

 if ( print > 0 ){
  cat("\n\nProcrustes Rotation with a Prescribed Factor Correlations\n")
  cat("\n This function finds the factor rotation matrix T of the form:")
  cat("\n   g=T'f and B=A inv(T')")
  cat("\n    where T = P inv(K) Q D R'")
  cat(",  corr(f)=Phia=P K2 P'  and  corr(g)=Phib=R D2 R'. \n")
  cat(" # of iterations required for convergence =", iter
      , " with eps=", eps, "  and epsd=", epsd,"\n")
  cat("  maxiter2=", maxiter2,", with eps2=", eps2, " and epsd2=", epsd2,"\n")
  cat(" SQUAREM=", SQUAREM,",  nSQUAREM=", nSQUAREM
      , " minalpha=", minalpha,",  maxalpha=", maxalpha
      , " reset1=", reset1,",  reset2=", reset2, "\n")
  cat("\n Final RMSE = ", rmse, "\n")
  cat("\n");
  Print(Q, T, fmt="7.3");
 }

 res=named_list( B, T, Q, Cm, A, C, Phia, Phib, rmse )
 return( res )

} # end of procrustes



