#' Procscal: Simultaneous Expanded Orthogonal Rotation of Many Matrices
#'
#' @param X Input array
#' @param mug Input variable mean matrix
#' @param estm = 1 to estimate m
#' @param model = 1 to set all W_k matrices proportional to identity matrix.\cr
#          = 2 to fit INDSCAL model: W_k is diagonal. \cr
#          = 3 to fit IDIOSCAL model    # not yet available
#' @param minW minimum value of W
#' @param maxiter max # of iterations
#' @param eps convergence criterion for rmse
#' @param maxiter2 max # of inner iterations
#' @param eps2 convergence criterion for inner iterations
#' @param epsd  convergence criterion for the difference
#' @param Xonly = 1 to use SQUAREM for X only.
#'
#' @param SQUAREM = 3 :   See the help of iSQUAREM in lazy.accel package.
#' @param nSQUAREM = 1 :   See the help of iSQUAREM in lazy.accel package.
#' @param minalpha = -999 :   See the help of iSQUAREM in lazy.accel package.
#' @param  maxalpha = -1 :   See the help of iSQUAREM in lazy.accel package.
#' @param  always = 1 :   See the help of iSQUAREM in lazy.accel package.
#' @param  reset1 = 0 :   See the help of iSQUAREM in lazy.accel package.
#' @param  reset2 = 1 :   See the help of iSQUAREM in lazy.accel package.

#' @param print = 1 to print the result
#' @param init = 1  to use pc of super matrix as the initial
#'
#' @param initpc = 0 to use random initial of princ
#' @param Xinit n x nr matrix consisting of the initial value of X matrix
#' @param Tinit nr x nr x nG array consisting of the rotation matrix
#' @param Winit nG x nr matrix consisting of the initial of
#'  the diag elem of W_k
#' @param minit nG x nr matrix consisting of the initial value of m
#'
#' @details
#' The program minimizes \cr
#'    RSS = sum_k ssq( ( X_k - (X W_k T_k + 1 m_k') ) ) \cr
#'  w.r.t X, T_k, m_k and W_k subject to T_k'T_k = T_k T_k' = I and W=diag. \cr
#' \cr
#'  In addition, if mug is given, instead of estimating m_k,
#' factor.mean will be calculated after conversion.
#'  \cr
#'  \cr
#'  Note on the iSQUAREM: \cr
#'  Try always=1 with maxalpha=-1 or less first. \cr
#'  Changing to reset1=1 and reset2=2 or increasing nSQUAREM may help. \cr
#'  If it seems not working, use always=0 with maxalpha=1 or less. \cr
#'  Changing to reset1=1 and reset2=2 may help. \cr
#'  If all of the above fail, be patient and use SQUAREM=0.
#'  \cr
#'
#' @return
#' When mu is present, a list of \cr
#'  X, W, T, m, q, u, mu, rmse \cr
#' where X, W, m, mu, u, and q are  group x dimension matrices
#' , T is a array. \cr
#' When mu is not present, u, q and mu are not included.
#'
#'
#' @examples
#' seed <- 1701
#' set.seed(seed)
#' nG <- 3; n <- 30; nr <- 2
#'
#' # Simultaneous rotation of several matirces
#' # generate data
#' resg <- gendataWmm( nG,n,nr, genm=1, genW=2, errstd=.05, pcA=1, minWD=0.4 )
#' Xall <- l2a(resg$C); mall <- t(a2m(l2a(resg$m),method=1))
#' Xg <- resg$A; Tg <- l2a(resg$T); Wg <- l2a(resg$W);
#' Wg <- t(matrix(colSums(a2m(Wg)),nr))
#' od <- nr+1-order( colSums(Wg*Wg) )
#' Wg <- Wg[,od];  Xg <- Xg[,od];  mg <- mall[,od]
#' for( k in 1:nG ){
#'  Tg[,,k] <- Tg[od,od,k];  Xall[,,k] <- Xall[,od,k]
#' }
#'
#' # Increase maxiter to obtain convergence!
#' res <- procscal( Xall, estm=1, model=2, minW=1e-4, init=1
#'               , always=1
#'               , maxiter=100, maxiter2=5, eps=1e-5, epsd=1e-4, print=2 )
#' X <- res$X; W <- res$W; T <- res$T; m <- res$m
#'
#'
#'
#' # Estimation of factor score means
#' seed <- 1701
#' set.seed(seed)
#' ncommonitems <- 10
#'
#' # generate data
#' resg <- gendataWmm( nG,n,nr, genmu=1,genW=2,errstd=0.05, pcA=1,minWD=0.4 )
#' Xall <- l2a(resg$C)
#' Xg <- resg$A; Tg <- l2a(resg$T); Wg <- l2a(resg$W);
#' mug <- a2m(l2a(resg$mu)); ug <- t( a2m(l2a(resg$u)) )
#' Wg <- t(matrix(colSums(a2m(Wg)),nr))
#' od <- nr+1-order( colSums(Wg*Wg) )
#' Wg <- Wg[,od]; Xg <- Xg[,od]; ug <- ug[,od]
#' for( k in 1:nG ){
#'  Tg[,,k] <- Tg[od,,k]; Xall[,,k] <- Xall[,od,k]
#' }
#'
#'
#' # Distribute items to nG forms:  systematic missing pattern
#' locmiss=gensysmiss( n, nG, ncommonitems, print=0, sort=0 )
#' for( k in 1:nG ){
#'  lmk <- locmiss[[k]]
#'  Xall[lmk,,k] <- NA; mug[lmk,k] <- NA
#' }
#' Print("Data to be analyzed:", Xall, mug, fmt="6.3")#'
#'
#' # Increase maxiter and lower eps values to obtain true convergence!
#' res=procscal( Xall, mug=mug, estm=0, model=2, minW=1e-4, init=1, always=1
#'             , maxiter=100, maxiter2=5, eps=1e-5, epsd=1e-4, print=1 )
#' X <- res$X; W <- res$W; T <- res$T; mu <- res$mu; u <- res$u; q <- res$q
#'
#'
#' @export
#'

procscal <- function( X, mug=NULL, estm=0, model=2, minW=1e-3
                    , maxiter=200, eps=1e-6, maxiter2=20, eps2=1e-5, epsd=1e-6
                    , Xonly=1
                    , SQUAREM=3, nSQUAREM=1
                    , minalpha=-999, maxalpha=-1, always=0
                    , reset1=0, reset2=1
                    , print=2, init=1, initpc=2
                    , Xinit=NULL, Tinit=NULL, Winit=NULL, minit=NULL ){
 # Expanded Orthogonal Procrustes Rotation
 # Shin-ichi Mayekawa
 # 111220,21
 # missing target: 111222
 # normalize Xc so that t(Xc)%*%Xc=diag when estW=0: 120104
 # procoteWm renamed as procscal
 # C -> X, A -> Xc and return nGxnr W: 120109
 # getrmse, X[locmiss]=1 : 120113
 # X[,j,k]=NA : 120114
 # Wall -> W: array to matrix: 120116
 # initial values: 120116
 # idioscal: 120119
 # estm==1 corrected: 120312c
 # estimation of factor means: 120319
 # q and R: 120322cj
 # bugfix: 121020
 # estimation of u and mu with missing elements: 121020,21
 # missing values for model=3: 121021
 # model=3 removed: 130124
 # minW for procoteW: 130124
 # pc initial with init=1: 120125
 # remove solo items: 130126,27
 # den of rmse: 130130
 # getsm -> getSM: 20150607
 # maxadA and maxadT -> abs(abs-abs): 20151221
 # iSQUAREM: 20151221,27, 20160104
 # lmax -> maxiter: 20170530dnc
 #
 #
 # Args:
 #  X       n x nr x nG array of target matrices
 #           X [,,k]=X_k, k=1,2,...,nG,   where X_k = n x nr
 #  mug     n x nG matrix of group means
 #          When mug != NULL factor means will be estimated.
 #  estm    = 1 to estimate m when mug is not present.
 #  model   = 1 to set T=W=I
 #          = 2 to fit INDSCAL model
 #          = 3 to fit IDIOSCAL model  # not yet available
 #  init    = 1 to use pc of super matrix as the initial
 #          = 0 to use default initial
 #            Actually when init=1, the better one is used.
 #  initpc  = 0 to use random initial of princ
 #  maxiter    max # of iterations
 #  eps     eps for the relative improvement of RMSE
 #  maxiter2   max # of the inner iterations
 #  eps2    eps for the relative improvement of RMSE
 #  print   to control the print level
 #
 #  Xinit   n x nr matrix consisting of the initial value of Xc matrix
 #  Winit   nG x nr matrix consisting of the initial of the diag elem of W_k
 #  Tinit   n x nr matrix consisting of the rotation matrix
 #  minit   n x nr matrix consisting of the initial value of mean
 #
 #
 # The program minimizes
 #    RSS = sum_k ssq( ( X_k - (Xc W_k T_k + 1 m_k') ) )
 #  w.r.t Xc, T_k, m_k and W_k subject to T_k'T_k = T_k T_k' = I and W=diag.
 #
 #  In addition, if mug is given, factor.mean will be calculated
 #  after conversion.
 #
 #
 #  In addition, if the n x 1 mean vectors for each group (mug) are given
 #  the factor score means for each group and the grand mean vector
 #  will be estimated assuming mean of the factor score means are zero.
 #
 #
 # The missing elements of X matrix will be estimated so that they also
 # minimize RSS.
 #
 # rmse = sqrt(RSS/(nG*n*nr)) will be reported.
 #
 #
 # Value: as list
 #   Xc     n x nr
 #   W      nG x nr matrix consisting of the diag elements of W_k
 #   T      nr x nr x nG array
 #   m      nG x nr
 #   rmse
 #
 #
 # Normalization of Xc matrix
 #  When estm  = 1, Xc will be colum centered.
 #       model = 1 or 3, Xc will be normalized so that t(Xc)*Xc=diag
 #       model = 2, Xc will be normalized so that Diag(t(Xc)%*Xc) = n I
 #
 #
 # Need:
 #  procoteW in procoteW1.R
 #

 getrmse <- function( title, nG, n, nr, X, Xc, W, Tall, mall
                    , rmsep, nobs, print=1 ){
  # for the debuging and initial only
  # seems that onen is from the parent frame.
   rss=0;
   for( k in 1:nG ){
    rss=rss+sum( ( X[,,k]-(Xc%*%Diag(W[k,])%*%Tall[,,k]+onen%*%mall[k,]) )^2 )
   }
   rmse=sqrt(rss/nobs)
   if( print > 0 ) cat(title, ":", rmse, "\n")
   if( rmsep < rmse ){
    cat(" ",title, ": **** failure !! ****", rmsep - rmse, " \n")
   }
   return( rmse )

 } # end of getrmse


 getmuu <- function( mug, A, print=0 ){
  # Given, mug and A, get mu and u such that
  #    mug[,g] = mu + A u[,g]
  # Shin-ichi Mayekawa
  # 20121020,21
  #
  # Note that, here, u[,g] is a column vector.
  #

  # const
  n=nrow(A); nr=ncol(A); nG=ncol(mug)

  # missing elements of vec(mug)
  locna=is.na(matrix(mug,,1))

  # non-full-rank design matrix
  X=cbind( rep(1,nG)%x%diag(n), diag(nG)%x%A )

  # delete the non observed rows
  X=X[!locna,]

  # one solution (not unique)
  muu=matSwp(t(X)%*%X)%*%t(X)%*%(matrix(mug,,1)[!locna])
  # muu=ginv(t(X)%*%X)%*%t(X)%*%(matrix(mug,,1)[!locna])
  mu=matrix(muu[1:n],n)
  u=matrix(muu[-(1:n)],nr,nG)

  # enforce the constraint that sum of u is equal to zero.
  ut=matrix(rowMeans(u),,1)
  u=u-ut%*%rep(1,,nG)
  # compensate by absorbing in mu
  mu=mu+A%*%ut

  if( print > 0 ){
   Print("getmuu", n,nr,nG)
   Print(mug, A)
   Print(mu,u)
  }

  return( list(mu=mu, u=u) )

 } # getmuu



 getrmse2 <- function( param ){
  # for iSQUAREM

  # Print("in getrmse2", param)
  if( Xonly == 0 ){
   if( estW == 0 ) Xc=matrix(param,n,nr)
   else if( estW == 2 ){
    Xc=matrix(param[1:(n*nr)],n,nr)
    W=matrix(param[-(1:(n*nr))],nG,nr)
   }
  }
  else Xc=matrix(param,n,nr)

  rss=0
  for( k in 1:nG ){
   rss=rss+sum( ( X[,,k]-(Xc%*%Diag(W[k,])%*%Tall[,,k]+onen%*%mall[k,]) )^2 )
  }
  rmse=sqrt(rss/nobs)
  return( rmse )
 } # end of getrmse2


 const <- function( param ){
  # for iSQUAREM
  if( Xonly == 0  &  estW == 2 ){
   W=matrix(param[-(1:(n*nr))],nG,nr)
   W[W < minW]=minW
   param[(n*nr+1):(nr*(n+nG))]=c(W)
  }
  return( param )
 } # end of const





 # original matrix
 Xorg=X
 n0=nrow(X)
 Xrname0=rownames(X)
 Xcname0=colnames(X)

 # missing target
 locmiss=is.na(X) # This is an array of the same size as X.
 nmiss0=sum(locmiss); pmiss0=nmiss0/(nG*n*nr)

 # solo and common items
 locsolo=which( apply(1-locmiss[,1,],1,sum) <= 1 )
 loccommon=which( apply(1-locmiss[,1,],1,sum) > 1 )
 nsolo=length(locsolo)
 ncommon=n0-nsolo  # this should be equal to n

 nobs=sum(!is.na(X))
 if( nsolo > 1 ){
  # pick up the common items only
  Xsolo=X[locsolo,,]
  X=X[loccommon,,]
  nobs=sum(!is.na(X))
 }
 locmiss=is.na(X) # This is an array of the same size as X.
 X[locmiss]=0



 # const
 nG=dim(X)[3]
 n=dim(X)[1]
 nr=dim(X)[2]
 nmiss=sum(locmiss); pmiss=nmiss/(nG*n*nr)


# Print(n0,nsolo,ncommon,n)

 onen=matrix(1,n,1)

 est_f.m=0
 if( !is.null(mug) ){
  est_f.m=1
  estm=0
 }

 # model
 if( model == 1 ) estW=0
 else if( model == 2 ) estW=2
 else if( model == 3 ) estW=0
 modelname=c("Identity", "INDSCAL", "IDIOSCAL")


 # initial
 Xgiven=0
 if( !is.null(Xinit) ){
  Xgiven=1
  Xc=Xinit
 }
 else{
  # group mean
  Xc=apply(X,c(1,2),mean, na.rm=1)
 }
 Tgiven=0
 if( !is.null(Tinit) ){
  Tgiven=1
  Tall=Tinit
 }
 else {
  # identity
  Tall=array(0,c(nr,nr,nG))
  for( k in 1:nG ){
   Tall[,,k]=diag(nr)
   }
 }
 Wgiven=0
 if( !is.null(Winit) ){
  Wgiven=1
  W=Winit
 }
 else{
  # all one
  W=matrix(1,nG,nr)
 }
 mgiven=0
 if( !is.null(minit) ){
  mgiven=1
  mall=minit
 }
 else{
  # column means of X
  mall=matrix(0,nG,nr)
  if( estm == 1 ){
   for( k in 1:nG )
    mall[k,]=matrix(colMeans(X[,,k], na.rm=1),1)
  }
 }
 # missing values
 for( k in 1:nG ){
  X[,,k][locmiss[,,k]]=(Xc%*%Diag(W[k,])%*%Tall[,,k])[locmiss[,,k]]
 }
 # initial rmse
 rmse_init0=getrmse( "", nG, n, nr, X, Xc, W, Tall, mall
                   , 99999, nobs=nobs, print=0 )
 rmse_init=rmse_init0
 Xcp=Xc; Tallp=Tall; Wp=W; mallp=mall
 init0=init

 if( init == 1 ){
  # pca of horizontally stacked super matrix
  temp=princ( a2m(X), nr, 0, print=0, init=initpc )
  Xc=temp$F
  tt=t( temp$A )
  hat=temp$Yhat
  for( k in 1:nG ){
  # Print(getsm(tt,k,1,nG),getSM(tt,1,k,SMdim=list(nrows=nr,ncols=rep(nr,nG))))
   # temp=svd(getsm(tt,k,1,nG))
   temp=svd(getSM(tt,1,k,SMdim=list(nrows=nr,ncols=rep(nr,nG))))
   Tall[,,k]=temp$u%*%t(temp$v)
  #Print(getsm(hat,k,1,nG),getSM(hat,1,k,SMdim=list(nrows=n,ncols=rep(nr,nG))))
   # X[,,k][locmiss[,,k]]=getsm(hat,k,1,nG)[locmiss[,,k]]
   X[,,k][locmiss[,,k]]=
       getSM(hat,1,k,SMdim=list(nrows=n,ncols=rep(nr,nG)))[locmiss[,,k]]
   W[k,]=mean(temp$d)
  }
  rm(temp,tt,hat)
  rmse_init1=getrmse( "", nG, n, nr, X, Xc, W, Tall, mall
                    , 99999, nobs=nobs, print=0 )
  #Print(rmse_init0, rmse_init1)
  # comparizon
  rmse_init=rmse_init1
  if( rmse_init0 < rmse_init1 ){
   # recover default initial
   init0=0
   rmse_init=rmse_init0
   Xc=Xcp; W=Wp; Tall=Tallp; mall=mallp
   for( k in 1:nG ){
    X[,,k][locmiss[,,k]]=(Xc%*%Diag(W[k,])%*%Tall[,,k])[locmiss[,,k]]
   }
  }
 }

 rownames(Xc)=rownames(X)
 colnames(Xc)=colnames(X)
 rownames(Tall)=colnames(X)
 colnames(Tall)=colnames(X)
 rownames(W)=paste("grp",1:nG,sep="")
 colnames(W)=colnames(X)
 dimnames(mall)=dimnames(W)


 # previous values
 Xcp=Xc; Tallp=Tall; Wp=W; mallp=mall
 rmse=rmse_init
 rmsep=rmse_init
 converged=0
 rmse0=99999999999

 if ( print > 0 ){
  cat("\n\n\nGeneralized Expanded Procrustes Rotation\n")
  cat("  minimizing\n")
  cat("   RSS = sum_k ssq(  X_k - 1 m_k' - Xc W_k T_k  ) \n\n")
  cat(" # of groups =", nG, ",  # of variables =", n0
    , ",  # of dimensions =", nr, "\n")
  cat(" # of missing elements in X = ", nmiss0, "(", pmiss0, ")\n")
  cat(" # of solo variables to be excluded =", nsolo,"\n")
  cat(" # of variables to be used for the estimation =", n,"\n")
  cat(" actual # of missing elements = ", nmiss, "(", pmiss, ")\n")
  cat(" actual # of observed elements (denominator of rmse) =", nobs,"\n")
  cat("\n estimation of m = ", estm
    , ",  estimation of factor means =", est_f.m,"\n")
  cat(" model = ", model, "(",
      modelname[model], ")\n")
  cat(" minimum value for W =", minW, "\n")
  cat(" initial values: Xc = ", Xgiven, ", W = ", Wgiven, ", T = ", Tgiven,
      ", m = ", mgiven, ",\n")
  cat(" init = ", init," (",init0, ")  with rmse =", rmse_init, "\n")
  cat("\n max # of iterations = ", maxiter, ", eps for relative rmseimp = "
      , eps, "\n")
  cat(" max # of iterations (procoteW) = ", maxiter2
      , ", eps for relative rmseimp = ", eps2, "\n")
  cat(" print = ", print, "\n")
 }


 ########################################################################
 # iSQUAREM
 if( Xonly == 0 ){
  if( estW == 0 ) param=c(Xc)
  else if( estW == 2 ) param=c(c(Xc),c(W))
 }
 else{
  param=c(Xc)
 }
 iSQUAREM=generate_iSQUAREM( param )
 ########################################################################


 # main iteration
 for( llll in 1:maxiter ){

  BB=0; CB=0
  for( k in 1:nG ){

   # estimate T and W
   TWm=procoteW( Xc, X[,,k], estm=estm, estW=estW, print=print-3
               , T=Tall[,,k], W=Diag(W[k,]), m=t(mall[k,,drop=F])
               , maxiter=maxiter2, eps=eps2, minW=minW )
   # store
   Tall[,,k]=TWm$T
   W[k,]=t(vecdiag(TWm$W))
   mall[k,]=t(TWm$m)

  if( print >= 5 ){
   tt=paste("after T & W for k =",k)
   rmse0=getrmse( tt, nG, n, nr, X, Xc, W, Tall, mall, rmse0, nobs=nobs )
  }

   # prepare for the estimation of Xc
   Bk=Diag(W[k,])%*%Tall[,,k]
   BB=BB+Bk%*%t(Bk)
   CB=CB+(X[,,k]-onen%*%mall[k,,drop=F])%*%t(Bk)

  } # end of k loop

  if( print >= 4 )
   rmse0=getrmse( "after T & W", nG, n, nr, X, Xc, W, Tall
                , mall, rmse0, nobs=nobs )


  # estimate Xc
  Xc=CB%*%solve(BB)

  if( print >= 4 )
   rmse0=getrmse( "after Xc", nG, n, nr, X, Xc, W, Tall, mall
                , rmse0, nobs=nobs )

  # Print("before SQEM", W, Xc, Tall, fmt="6.3")
  ###################################################################
  if( SQUAREM > 0  &&  nSQUAREM <= llll ){
   if( Xonly == 0 ){
    if( estW == 0 ) param=c(Xc)
    else if( estW == 2 ) param=c(c(Xc),c(W))
   }
   else{
    param=c(Xc)
   }

   temp=iSQUAREM( param
                 , enforce_constraints=const, badness_of_fit=getrmse2
                 , SQUAREM=SQUAREM, minalpha=minalpha, maxalpha=maxalpha
                 , bof_value=NULL, always=always, reset1=reset1, reset2=reset2
                 , print=max(0,print-1), debug=0 )
   rmse0=temp$critval
   param=temp$param
   if( Xonly == 0 ){
    if( estW == 0 ) Xc=matrix(param,n,nr)
    else if( estW == 2 ){
     Xc=matrix(param[1:(n*nr)],n,nr)
     W=matrix(param[-(1:(n*nr))],nG,nr)
    }
   }
   else{
    Xc=matrix(param,n,nr)
   }
  }
  ###################################################################
  # Print("after SQEM", W, Xc, Tall, fmt="6.3")


  # normalize Xc and absorb in W or T
  if( estm == 1 ){
   mA=colMeans(Xc)
   Xc=Xc-onen%*%t(mA)
   for( k in 1:nG ){
    mall[k,]=mall[k,] + t(Tall[,,k])%*%Diag(W[k,])%*%mA
   }
   if( print >= 4 )
    rmse0=getrmse( "after norm Xc", nG, n, nr, X, Xc, W, Tall
                 , mall, rmse0, nobs=nobs )
  }

  if( estW == 2 ){
   # equal column ssq
   DA=Diag(sqrt(colSums(Xc*Xc)/n))
   DAinv=solve(DA)
   Xc=Xc%*%DAinv
   for( k in 1:nG ){
    W[k,]=W[k,]%*%DA
   }
  } else if( estW == 0 ){
   # principal component
   ev=eigen(t(Xc)%*%Xc);
   eva=matrix(ev$values,length(ev$values),1); eve=ev$vectors;
   Xc=Xc%*%eve
   for( k in 1:nG ){
    Tall[,,k]=t(eve)%*%Tall[,,k]
   }
  }

  # check convergenge
  rss=0;
  for( k in 1:nG ){
   rss=rss+
    sum( ( X[,,k]-(Xc%*%Diag(W[k,])%*%Tall[,,k]+onen%*%mall[k,,drop=F]) )^2 )
  }
  rmse=sqrt(rss/nobs)
  rmseimp=rmsep-rmse
  rmseimpr=rmseimp/(rmsep+1e-7)
  maxadA=max(abs(abs(Xc)-abs(Xcp)))
  maxadT=max(abs(abs(Tall)-abs(Tallp)))

  if ( print >= 2 ){
   Print(llll, rmse, rmsep, rmseimpr, maxadA, maxadT, fmt="i2. 8.5")
  }

  if ( rmse < 1e-7 || (rmseimpr < eps && maxadA < epsd && maxadT < epsd) ){
   converged=1;
   break;
  }


  # update missing elements in X
  if( pmiss > 0 ){
   for( k in 1:nG ){
    lmk=locmiss[,,k]
    if( any(lmk) ){
     B=Xc%*%Diag(W[k,])%*%Tall[,,k]+onen%*%mall[k,,drop=F]
     Ck=X[,,k]; Ck[lmk]=B[lmk];
     X[,,k]=Ck
    }
   }
   if( print >= 4 )
    rmse0=getrmse( "after missing", nG, n, nr, X, Xc, W, Tall
                 , mall, rmse0, nobs=nobs )
  }


  # next iteration;
  Xcp=Xc;
  Tallp=Tall
  rmsep=rmse;


 } # end of llll loop


 if ( print >= 1 )
  Print(llll, rmse, rmseimpr, maxadA, maxadT, fmt="i4 10.7");


 iter=llll



 colnames(Xc)=colnames(X)

 # recover solo items
 if( length(locsolo) > 0 ){
  Xc1=matrix(NA,n0,nr)
  rownames(Xc1)=Xrname0
  Xc1[loccommon,]=Xc
  for( k in 1:nG ){
   loc=which( !is.na(Xsolo[,1,k]) )
   if( length(loc) > 0 ){
    locloc=locsolo[loc]
#    Print(k, Xsolo[,,k],loc, locloc)
    Xc2=Xsolo[loc,,k]%*%solve(W[k,]*Tall[,,k])
    Xc1[locloc,]=Xc2
   }
  }
 Xc=Xc1
 }
 X=Xorg

 dimnames(Xc)=list(Xrname0,Xcname0)


 # normalize Xc and absorb in W or T
 onen=matrix(1,n0,1)
 if( estm == 1 ){
  mA=matrix(colSums(Xc)/n,nr)
  Xc=Xc-onen%*%t(mA)
  for( k in 1:nG ){
   mall[k,]=mall[k,]+t( t(Tall[,,k])%*%Diag(W[k,])%*%mA )
  }
  if( print >= 4 )
   rmse0=getrmse( "after norm Xc", nG, n, nr, X, Xc, W, Tall
                , mall, rmse0, nobs=nobs )
 }

 if( estW == 2 ){
  # equal column ssq
  DA=Diag(sqrt(colSums(Xc*Xc)/n0))
  DAinv=solve(DA)
  Xc=Xc%*%DAinv
  for( k in 1:nG ){
   W[k,]=W[k,]%*%DA
  }
 } else if( estW == 0 ){
  # principal component
  ev=eigen(t(Xc)%*%Xc);
  eva=matrix(ev$values,length(ev$values),1); eve=ev$vectors;
  Xc=Xc%*%eve
  for( k in 1:nG ){
   Tall[,,k]=t(eve)%*%Tall[,,k]
  }
 }
 ########################################################################

 colnames(Xc)=colnames(X)


 # sort the dimensions according to W^2
 od=nr+1-order( colSums(W*W) )
 W=W[,od]
 Xc=Xc[,od]
 # mall=mall[,od]
 for( k in 1:nG ){
  Tall[,,k]=Tall[od,,k]
 }

 tXcXc=t(Xc)%*%Xc / n0


 # factor score means
 ug=matrix(0,nG);
 qg=ug
 if( est_f.m == 1 ){
  resmuu=getmuu( mug, Xc )
  ug=resmuu$u
  mu=resmuu$mu
  R=1/W
  colnames(ug)=rownames(W)
  rownames(ug)=colnames(W)
  qg=ug
  for( g in 1:nG ){
   qg[,g]=-diag(R[g,])%*%ug[,g]
  }

  ################ transpose the results ################## 130127
  ug=t(ug); qg=t(qg)
 }


 if ( print > 0 ){
  cat("Iteration Terminated.\n")
  Print(llll, rmse, rmseimpr, maxadA, maxadT, fmt="i4 10.7");
  cat("\n\n\nGeneralized Expanded Procrustes Rotation\n")
  cat("  minimizing\n")
  cat("   RSS = sum_k ssq(  X_k - 1 m_k' - Xc W_k T_k  ) \n\n")
  cat(" # of groups =", nG, ",  # of variables =", n0
   , ",  # of dimensions =", nr, "\n")
  cat(" # of missing elements in X = ", nmiss0, "(", pmiss0, ")\n")
  cat(" # of solo variables to be excluded =", nsolo,"\n")
  cat(" # of variables to be used for the estimation =", n,"\n")
  cat(" actual # of missing elements = ", nmiss, "(", pmiss, ")\n")
  cat(" actual # of observed elements (denominator of rmse) =", nobs,"\n")
  cat("\n estimation of m = ", estm
   , ",  estimation of factor means =", est_f.m,"\n")
  cat(" model = ", model, "(",
   modelname[model], ")\n")
  cat(" minimum value for W =", minW, "\n")
  cat(" initial values: Xc = ", Xgiven, ", W = ", Wgiven, ", T = ", Tgiven,
   ", m = ", mgiven, ",\n")
  cat(" init = ", init," (",init0, ")  with rmse =", rmse_init, "\n")
  cat("\n max # of iterations = ", maxiter, ", eps for relative rmseimp = "
      , eps, "\n")
  cat(" max # of iterations (procoteW) = ", maxiter2
   , ", eps for relative rmseimp = ", eps2, "\n")
  cat(" # of iterations required =", iter, "\n")
  cat(" Final RMSE = ", rmse,",  maxadT =", maxadT
    , ",  maxadA =",  maxadA, "\n")
  cat(" print = ", print, "\n")
  cat("\n");
  Print(Xc, fmt="10.4");
  Print(tXcXc, fmt="10.4")
  Print(W, fmt="10.4");
  Print(Tall, fmt="10.4");
  Print(mall, fmt="10.4");
  if( est_f.m == 1 ){
   cat(" Factor Means for each group (qg)\n")
   Print(ug,qg, fmt="10.4")
   cat("\n grand mean\n")
   Print(mu, fmt="10.4")
  }
 }

 # returning the result
 if( est_f.m == 1 )
  res=list(X=Xc, W=W, T=Tall, m=mall, q=qg, u=ug, mu=mu, rmse=rmse)
 else
  res=list(X=Xc, W=W, T=Tall, m=mall, rmse=rmse)

 return( res )


} # end of procscal


